# pylint: skip-file
import pyomo.environ as pe
import pytest
from hypothesis import given, assume
from hypothesis.strategies import integers, lists
from pyomo.core.kernel.component_map import ComponentMap

from suspect.fbbt.propagation.rules import *
from suspect.fbbt.propagation.visitor import BoundsPropagationVisitor
from suspect.interval import Interval
from suspect.pyomo.expressions import *
from tests.strategies import expressions, reals, intervals


@pytest.fixture
def visitor():
    return BoundsPropagationVisitor()


@given(reals(), reals())
def test_variable_bound(visitor, a, b):
    lb = min(a, b)
    ub = max(a, b)
    var = pe.Var(bounds=(lb, ub))
    var.construct()
    matched, result = visitor.visit_expression(var, None)
    assert matched
    assert result == Interval(lb, ub)


@given(reals())
def test_constant_bound(visitor, c):
    const = NumericConstant(c)
    matched, result = visitor.visit_expression(const, None)
    assert matched
    assert result == Interval(c, c)


@given(reals())
def test_constant_of_float_bound(visitor, const):
    rule = ConstantRule()
    matched, result = visitor.visit_expression(const, None)
    assert result == Interval(const, const)


@given(reals(), reals(), reals(), reals())
def test_constraint_bound(visitor, a, b, c, d):
    e_lb, e_ub = min(a, b), max(a, b)
    c_lb, c_ub = min(c, d), max(c, d)
    assume(max(e_lb, c_lb) <= min(e_ub, c_ub))
    child = pe.Var()

    cons = Constraint('c0', lower_bound=c_lb, upper_bound=c_ub, children=[child])
    bounds = ComponentMap()
    bounds[child] = Interval(e_lb, e_ub)
    matched, result = visitor.visit_expression(cons, bounds)
    expected = Interval(max(e_lb, c_lb), min(c_ub, e_ub))
    assert result == expected


@given(reals(max_value=100.0), reals(max_value=100.0))
def test_objective_bound(visitor, a, b):
    lb, ub = min(a, b), max(a, b)
    assume(lb < ub)
    child = pe.Var()
    o = Objective('obj', children=[child])
    bounds = ComponentMap()
    bounds[child] = Interval(lb, ub)
    matched, result = visitor.visit_expression(o, bounds)
    assert matched
    assert result == Interval(lb, ub)


# pyomo has binary products
@given(lists(intervals(), min_size=2, max_size=2))
def test_product_visitor(visitor, children_bounds):
    children = [pe.Var() for _ in children_bounds]
    bounds = ComponentMap()
    expected = 1.0
    for child, child_bound in zip(children, children_bounds):
        bounds[child] = child_bound
        expected *= child_bound

    expr = children[0] * children[1]
    matched, result = visitor.visit_expression(expr, bounds)
    assert matched
    assert result == expected


@pytest.mark.parametrize('bound,expected', [
    (Interval(-1, 1), Interval(None, None)),
    (Interval(2, 10), Interval(0.1, 0.5)),
])
def test_division_bound(visitor, bound, expected):
    num = pe.Var()
    child = pe.Var()
    expr = num / child
    bounds = ComponentMap()
    bounds[num] = Interval(1.0, 1.0)
    bounds[child] = bound
    matched, result = visitor.visit_expression(expr, bounds)
    assert matched
    assert result == expected


@given(expressions(), intervals())
def test_negation_bound(visitor, child, child_bounds):
    expr = -child
    assume(child.is_expression_type())
    assume(expr.is_expression_type() and not isinstance(expr, MonomialTermExpression))
    bounds = ComponentMap()
    bounds[child] = child_bounds
    matched, result = visitor.visit_expression(expr, bounds)
    assert matched
    assert -result == child_bounds


@given(expressions(), integers(min_value=1))
def test_even_pow_bound(visitor, base, n):
    c = NumericConstant(2*n)
    pow_ = base ** c
    bounds = ComponentMap()
    matched, result = visitor.visit_expression(pow_, bounds)
    assert matched
    assert result.is_nonnegative()


@given(expressions(), expressions())
def test_even_pow_non_const_bound(visitor, base, expo):
    assume(not expo.is_constant())
    expr = base ** expo
    bounds = ComponentMap()
    matched, result = visitor.visit_expression(expr, bounds)
    assert matched
    assert result == Interval(None, None)


@given(expressions(), integers(min_value=1))
def test_even_pow_odd_bound(visitor, base, n):
    c = 2 * n + 1
    expr = base ** c
    bounds = ComponentMap()
    matched, result = visitor.visit_expression(expr, bounds)
    assert matched
    assert result == Interval(None, None)


@given(expressions(), integers(min_value=1))
def test_even_pow_negative_bound(visitor, base, n):
    c = -2 * n
    expr = base ** c
    bounds = ComponentMap()
    matched, result = visitor.visit_expression(expr, bounds)
    assert matched
    assert result == Interval(None, None)


@pytest.mark.parametrize('func_name,interval,expected', [
    ('abs', Interval(-10, 3), Interval(0, 10)),
    ('sqrt', Interval(4, 16), Interval(2, 4)),
    ('exp', Interval(0, 1).log(), Interval(0, 1)),
    ('log', Interval(0, 1).exp(), Interval(0, 1)),
    ('sin', Interval(0, 1).asin(), Interval(0, 1)),
    ('cos', Interval(0, 1).acos(), Interval(0, 1)),
    ('tan', Interval(0, 1).atan(), Interval(0, 1)),
    ('asin', Interval(0, 1).sin(), Interval(0, 1)),
    ('acos', Interval(0, 1).cos(), Interval(0, 1)),
    ('atan', Interval(0, 1).tan(), Interval(0, 1)),
])
@given(child=expressions())
def test_unary_function(visitor, func_name, interval, expected, child):
    if func_name == 'abs':
        expr = abs(child)
    else:
        expr = getattr(pe, func_name)(child)
    bounds = ComponentMap()
    bounds[child] = interval
    matched, result = visitor.visit_expression(expr, bounds)
    assert matched
    assert result == expected


@given(expressions())
def test_visitor_tightens_new_bounds(visitor, expr):
    bounds = ComponentMap()
    assert bounds.get(expr, None) is None

    assert visitor.handle_result(expr, Interval(0, 2), bounds)
    assert bounds[expr] == Interval(0, 2)

    assert not visitor.handle_result(expr, Interval(0, 2), bounds)
    assert bounds[expr] == Interval(0, 2)

    assert visitor.handle_result(expr, Interval(0, 1), bounds)
    assert bounds[expr] == Interval(0, 1)



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Convexity detection rules for base expressions."""
from suspect.convexity.convexity import Convexity
from suspect.convexity.rules.rule import ConvexityRule


class VariableRule(ConvexityRule):
    """Return convexity of variable."""
    def apply(self, _expr, _convexity, _mono, _bounds):
        return Convexity.Linear


class ConstantRule(ConvexityRule):
    """Return convexity of constant."""
    def apply(self, _expr, _convexity, _mono, _bounds):
        return Convexity.Linear


class ConstraintRule(ConvexityRule):
    """Return convexity of constraint."""
    def apply(self, expr, convexity, _mono, _bounds):
        child = expr.args[0]
        cvx = convexity[child]
        if expr.bounded_below() and expr.bounded_above():
            # l <= g(x) <= u
            if cvx.is_linear():
                return cvx
            return Convexity.Unknown
        elif expr.bounded_below():
            # l <= g(x) => -g(x) <= -l
            return cvx.negate()
        elif expr.bounded_above():
            # g(x) <= u
            return cvx
        raise RuntimeError('Constraint with no bounds')  # pragma: no cover


class ObjectiveRule(ConvexityRule):
    """Return convexity of objective."""
    def apply(self, expr, convexity, _mono, _bounds):
        child = expr.args[0]
        cvx = convexity[child]
        if expr.is_minimizing():
            return cvx
        # max f(x) == min -f(x)
        return cvx.negate()



from typing import NamedTuple

import pyomo.environ as pyo
from pyomo.core.expr.numeric_expr import (
    SumExpression,
)
from pyomo.repn.standard_repn import generate_standard_repn


def enable_standard_repn_for_quadratic_expression():
    from pyomo.repn.standard_repn import _repn_collectors, _collect_sum
    _repn_collectors[QuadraticExpression] = _collect_sum


class BilinearTerm(NamedTuple):
    var1: pyo.Var
    var2: pyo.Var
    coefficient: float


def _make_uid_index(v0, v1):
    v0_uid = id(v0)
    v1_uid = id(v1)
    if v0_uid < v1_uid:
        return (v0, v0_uid), (v1, v1_uid)
    return (v1, v1_uid), (v0, v0_uid)


class QuadraticExpression(SumExpression):
    def __init__(self, args):
        if isinstance(args, SumExpression):
            args = args.args
        if not isinstance(args, list):
            args = [args]
        super().__init__(args)
        self._coef_matrix = None

    @property
    def __class__(self):
        return SumExpression

    def coefficient(self, v0, v1):
        self._ensure_coef_matrix()
        (_, v0_uid), (_, v1_uid) = _make_uid_index(v0, v1)
        bilinear_term = self._coef_matrix.get((v0_uid, v1_uid), None)
        if bilinear_term is None:
            return 0.0
        return bilinear_term.coef

    @property
    def terms(self):
        self._ensure_coef_matrix()
        return self._coef_matrix.values()

    def _ensure_coef_matrix(self):
        if not hasattr(self, '_coef_matrix'):
            self._coef_matrix = None

        if self._coef_matrix is not None:
            return

        self._coef_matrix = dict()

        repn_result = generate_standard_repn(self, compute_values=False)
        assert not repn_result.linear_vars
        assert not repn_result.nonlinear_vars

        for ((v0, v1), coef) in zip(repn_result.quadratic_vars, repn_result.quadratic_coefs):
            (v0, v0_uid), (v1, v1_uid) = _make_uid_index(v0, v1)
            assert v0_uid <= v1_uid
            bilinear_term = BilinearTerm(v0, v1, coef)
            self._coef_matrix[(v0_uid, v1_uid)] = bilinear_term

        if repn_result.nonlinear_expr is not None:
            pass
        else:
            assert not repn_result.nonlinear_vars



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Polynomiality rules."""
from suspect.pyomo.expressions import nonpyomo_leaf_types
from suspect.polynomial.degree import PolynomialDegree
from suspect.interfaces import Rule


class PolynomialDegreeRule(Rule):
    """Base class for rules that compute polynomial degree
     properties of an expression."""
    def apply(self, expr, polynomial_degree):
        """Apply rule to `expr`.

        Parameters
        ----------
        expr : Expression
            the expression
        polynomial_degree : dict-like
            contains polynomial degree of children
        """
        raise NotImplementedError('apply')


class VariableRule(PolynomialDegreeRule):
    """Return polynomial degree of variable."""
    def apply(self, _expr, _poly):
        return PolynomialDegree(1)


class ConstantRule(PolynomialDegreeRule):
    """Return polynomial degree of constant."""
    def apply(self, _expr, _poly):
        return PolynomialDegree(0)


class ConstraintRule(PolynomialDegreeRule):
    """Return polynomial degree of constraint."""
    def apply(self, expr, poly):
        return poly[expr.args[0]]


class ObjectiveRule(PolynomialDegreeRule):
    """Return polynomial degree of objective."""
    def apply(self, expr, poly):
        return poly[expr.args[0]]


class DivisionRule(PolynomialDegreeRule):
    """Return polynomial degree of division.

    If the denominator is constant, it returns the polynomial degree of the numerator,
    otherwise return no polynomial degree.
    """
    def apply(self, expr, poly):
        den_degree = poly[expr.args[1]]
        if den_degree.is_polynomial() and den_degree.degree == 0:
            # constant denominator
            return poly[expr.args[0]]
        return PolynomialDegree.not_polynomial()


class ReciprocalRule(PolynomialDegreeRule):
    """Return polynomial degree of reciprocal."""
    def apply(self, expr, poly):
        den_degree = poly[expr.args[0]]
        if den_degree.is_polynomial() and den_degree.degree == 0:
            return PolynomialDegree(0)
        return PolynomialDegree.not_polynomial()


class ProductRule(PolynomialDegreeRule):
    """Return polynomial degree of product.

    The polynomial degree of a product is equal to the sum of the polynomial degre of its children.
    """
    def apply(self, expr, poly):
        return sum([poly[a] for a in expr.args], PolynomialDegree(0))


class LinearRule(PolynomialDegreeRule):
    """Return polynomial degree of a linear expression."""
    def apply(self, expr, poly):
        if not expr.args:
            return PolynomialDegree(0)
        return PolynomialDegree(1)


class QuadraticRule(PolynomialDegreeRule):
    """Return polynomial degree of a quadratic expression."""
    def apply(self, expr, _poly):
        if not expr.args:
            return PolynomialDegree(0)
        return PolynomialDegree(2)


class SumRule(PolynomialDegreeRule):
    """Return polynomial degree of a summation expression."""
    def apply(self, expr, poly):
        return max(poly[a] for a in expr.args)


class NegationRule(PolynomialDegreeRule):
    """Return polynomial degree of a negation expression."""
    def apply(self, expr, poly):
        return poly[expr.args[0]]


class AbsRule(PolynomialDegreeRule):
    """Return polynomial degree of a abs expression."""
    def apply(self, expr, poly):
        return poly[expr.args[0]]


class PowerRule(PolynomialDegreeRule):
    """Return polynomial degree of a power expression."""
    def apply(self, expr, poly):
        assert len(expr.args) == 2
        base, expo = expr.args
        base_degree = poly[base]
        expo_degree = poly[expo]

        if not (base_degree.is_polynomial() and expo_degree.is_polynomial()):
            return PolynomialDegree.not_polynomial()

        if expo_degree.degree == 0:
            if base_degree.degree == 0:
                # const ** const
                return PolynomialDegree(0)
            if type(expo) not in nonpyomo_leaf_types:
                if not expo.is_constant():
                    return PolynomialDegree.not_polynomial()
                expo = expo.value
            if expo == int(expo):
                return base_degree ** expo
        return PolynomialDegree.not_polynomial()


class UnaryFunctionRule(PolynomialDegreeRule):
    """Return polynomial degree of an unary function."""
    def apply(self, expr, poly):
        return PolynomialDegree.not_polynomial()



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Monotonicity detection rules for base expressions."""
from suspect.monotonicity.monotonicity import Monotonicity
from suspect.expression import ExpressionType
from suspect.monotonicity.rules.rule import MonotonicityRule


class VariableRule(MonotonicityRule):
    """Return monotonicity of variable."""
    def apply(self, _expr, _mono, _bounds):
        return Monotonicity.Nondecreasing


class ConstantRule(MonotonicityRule):
    """Return monotonicity of constant."""
    def apply(self, _expr, _mono, _bounds):
        return Monotonicity.Constant


class ConstraintRule(MonotonicityRule):
    """Return monotonicity of constraint."""
    def apply(self, expr, monotonicity, _bounds):
        assert len(expr.args) == 1
        child = expr.args[0]
        mono = monotonicity[child]
        if expr.bounded_below() and expr.bounded_above():
            # l <= g(x) <= u
            if mono.is_constant():
                return mono
            return Monotonicity.Unknown
        elif expr.bounded_below():
            # l <= g(x) => -g(x) <= -l
            return mono.negate()
        elif expr.bounded_above():
            # g(x) <= u
            return mono
        raise RuntimeError('Constraint with no bounds')  # pragma: no cover


class ObjectiveRule(MonotonicityRule):
    """Return monotonicity of objective."""
    def apply(self, expr, monotonicity, bounds):
        assert len(expr.args) == 1
        child = expr.args[0]
        mono = monotonicity[child]
        if expr.is_minimizing():
            return mono
        # max f(x) == min -f(x)
        return mono.negate()



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# from suspect.monotonicity.monotonicity import Monotonicity
# from suspect.convexity.convexity import Convexity

"""Module with classes for extending SUSPECT."""
from suspect.visitor import ForwardVisitor


class ConvexityDetector(ForwardVisitor):
    """Convexity Detector base class."""
    needs_matching_rules = False

    def register_rules(self):
        raise NotImplementedError('register_rules')

    def handle_result(self, expr, result, ctx):
        if result is None:
            return False
        ctx.convexity[expr] = result
        return not result.is_unknown()



# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import pyomo.environ as aml
import hypothesis.strategies as st
from suspect.dag import *


@pytest.fixture
def model():
    m = aml.ConcreteModel()
    m.I = range(10)
    m.J = range(5)
    m.x = aml.Var(m.I)
    m.y = aml.Var(m.I, m.J, domain=aml.NonNegativeReals, bounds=(0, 1))
    m.z = aml.Var(m.I, bounds=(0, 10))
    return m


@st.composite
def variables(draw):
    domain = draw(st.sampled_from(
        Domain.REALS,
        Domain.INTEGERS,
        Domain.BINARY,
        ))
    lower_bound = draw(st.one_of(st.none(), st.floats()))
    upper_bound = draw(st.one_of(st.none(), st.floats(min_value=lower_bound)))
    return Variable(lower_bound, upper_bound, domain)



# pylint: skip-file
import pytest
from pathlib import Path
from suspect.pyomo.util import (
    model_variables,
    model_objectives,
    model_constraints,
)
from suspect.pyomo import read_osil
import pyomo.environ as pe
from pyomo.core.expr.sympy_tools import sympyify_expression


current_dir = Path(__file__).resolve().parent


def test_osil_read():
    filename = current_dir / 'osil' / 'example1.xml'
    problem = read_osil(str(filename))
    variables = list(model_variables(problem))
    constraints = list(model_constraints(problem))
    objectives = list(model_objectives(problem))
    assert len(variables) == 35
    assert len(constraints) == 18
    assert len(objectives) == 1

    m = problem

    e9_body_expected = m.x33 - (m.x32*pe.log(m.x22)/(1.0-m.x32))
    e9_body_actual = m.e9.body
    assert sympyify_expression(e9_body_expected - e9_body_actual)[1] == 0

    e11_body_expected = m.x35 - ((m.x33 - m.x34*(m.x32+1.0))/(m.x32+3.0)/(1.0-m.x32)**2/m.x34)
    e11_body_actual = m.e11.body
    assert sympyify_expression(e11_body_expected - e11_body_actual)[1] == 0

    e14_body_expected = m.x26 - m.x22 * m.x28
    e14_body_actual = m.e14.body
    assert sympyify_expression(e14_body_expected - e14_body_actual)[1] == 0

    e15_body_expected = m.x27 - m.x22 * m.x29
    e15_body_actual = m.e15.body
    assert sympyify_expression(e15_body_expected - e15_body_actual)[1] == 0



# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import pkg_resources
import pyomo.environ as pyo
from pyomo.core.kernel.component_map import ComponentMap
from pyomo.core.expr.visitor import StreamBasedExpressionVisitor
from suspect.monotonicity import MonotonicityPropagationVisitor, propagate_expression_monotonicity
from suspect.convexity import ConvexityPropagationVisitor, propagate_expression_convexity
from suspect.dag.iterator import DagForwardIterator


logger = logging.getLogger('suspect')


def monotonicity_detection_entry_points():
    return pkg_resources.iter_entry_points('suspect.monotonicity_detection')


def convexity_detection_entry_points():
    return pkg_resources.iter_entry_points('suspect.convexity_detection')


class SpecialStructurePropagationVisitor(object):
    def __init__(self, problem):
        self._mono_visitors = [MonotonicityPropagationVisitor()]
        for entry_point in monotonicity_detection_entry_points():
            cls = entry_point.load()
            self._mono_visitors.append(cls(problem))

        logger.info(
            'Loaded %s monotonicity detectors: %s',
            len(self._mono_visitors), ', '.join([str(mono) for mono in self._mono_visitors])
        )

        self._cvx_visitors = [ConvexityPropagationVisitor()]
        for entry_point in convexity_detection_entry_points():
            cls = entry_point.load()
            self._cvx_visitors.append(cls(problem))

        logger.info(
            'Loaded %s convexity detectors: %s',
            len(self._cvx_visitors), ', '.join([str(cvx) for cvx in self._cvx_visitors])
        )

    def visit(self, expr, convexity, mono, bounds):
        for mono_visitor in self._mono_visitors:
            mono_known = mono_visitor.visit(expr, mono, bounds)
            if mono_known:
                break

        for cvx_visitor in self._cvx_visitors:
            cvx_known = cvx_visitor.visit(expr, convexity, mono, bounds)
            if cvx_known:
                break
        return [expr]


def propagate_special_structure(model, bounds, active=True):
    """Propagate special structure.

    Arguments
    ---------
    problem: ProblemDag
        the problem.
    bounds: dict-like
        a dict-like object containing bounds
    active : bool
        only propagate special structure on active components

    Returns
    -------
    monotonicity: dict-like
        monotonicity information for the problem.
    convexity: dict-like
        convexity information for the problem.
    """
    mono = ComponentMap()
    cvx = ComponentMap()

    for objective in model.component_data_objects(pyo.Objective, active=active, descend_into=True):
        _propagate_special_structure(objective.expr, bounds, cvx, mono)

    for constraint in model.component_data_objects(pyo.Constraint, active=active, descend_into=True):
        _propagate_special_structure(constraint.body, bounds, cvx, mono)

    return mono, cvx


def _propagate_special_structure(expr, bounds, cvx, mono):
    def enter_node(node):
        return None, None

    def exit_node(node, data):
        mono_result = propagate_expression_monotonicity(node, mono, bounds)
        mono[node] = mono_result
        result = propagate_expression_convexity(node, cvx, mono, bounds)
        cvx[node] = result

    return StreamBasedExpressionVisitor(
        enterNode=enter_node,
        exitNode=exit_node,
    ).walk_expression(expr)



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""FBBT bounds tightening visitor."""

from suspect.fbbt.tightening.rules import (
    ConstraintRule,
    SumRule,
    LinearRule,
    PowerRule,
    MonomialTermRule,
    UnaryFunctionRule,
    QuadraticRule,
)
from suspect.interval import Interval
from suspect.pyomo.quadratic import QuadraticExpression
from suspect.pyomo.expressions import (
    nonpyomo_leaf_types,
    Constraint,
    LinearExpression,
    SumExpression,
    PowExpression,
    MonomialTermExpression,
    UnaryFunctionExpression,
)
from suspect.visitor import BackwardVisitor

_expr_to_rule_map = dict()
_expr_to_rule_map[Constraint] = ConstraintRule()
_expr_to_rule_map[LinearExpression] = LinearRule()
_expr_to_rule_map[SumExpression] = SumRule()
_expr_to_rule_map[PowExpression] = PowerRule()
_expr_to_rule_map[UnaryFunctionExpression] = UnaryFunctionRule()
_expr_to_rule_map[MonomialTermExpression] = MonomialTermRule()
_expr_to_rule_map[QuadraticExpression] = QuadraticRule()


def tighten_bounds_root_to_leaf(expr, bounds):
    if type(expr) in nonpyomo_leaf_types:
        return None
    if not expr.is_expression_type():
        return None
    rule = _expr_to_rule_map.get(type(expr), None)
    if rule is not None:
        return rule.apply(expr, bounds)
    return None


class BoundsTighteningVisitor(BackwardVisitor):
    """Tighten bounds from sinks to sources."""
    needs_matching_rules = False
    intersect_abs_eps = 1e-10

    def visit_expression(self, expr, bounds):
        new_bounds = tighten_bounds_root_to_leaf(expr, bounds)
        if new_bounds is not None:
            return True, new_bounds
        return False, None

    def handle_result(self, expr, value, bounds):
        if value is None:
            new_bounds = Interval(None, None)
            new_bounds = value

        old_bounds = bounds.get(expr, None)
        if old_bounds is not None:
            new_bounds = old_bounds.intersect(
                new_bounds,
                abs_eps=self.intersect_abs_eps,
            )
            has_changed = old_bounds != new_bounds
        else:
            has_changed = True
        bounds[expr] = new_bounds
        return has_changed



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Monotonicity detection rules for power expressions."""
from suspect.math import almosteq, \
    almostgte  # pylint: disable=no-name-in-module
from suspect.monotonicity.monotonicity import Monotonicity
from suspect.monotonicity.rules.rule import MonotonicityRule
from suspect.pyomo.expressions import nonpyomo_leaf_types


class PowerRule(MonotonicityRule):
    """Return monotonicity of power expression."""
    def apply(self, expr, monotonicity, bounds):
        base, expo = expr.args

        mono_base = monotonicity[base]
        mono_expo = monotonicity[expo]

        bounds_base = bounds[base]
        bounds_expo = bounds[expo]

        if mono_base.is_constant():
            if type(base) not in nonpyomo_leaf_types:
                base = base.value
            return _monotonicity_constant_base(
                base, expo,
                mono_base, mono_expo,
                bounds_base, bounds_expo,
            )
        elif mono_expo.is_constant():
            if type(expo) not in nonpyomo_leaf_types:
                expo = expo.value
            return _monotonicity_constant_exponent(
                base, expo,
                mono_base, mono_expo,
                bounds_base, bounds_expo,
            )
        return Monotonicity.Unknown


def _monotonicity_constant_base(base, _expo, _mono_base, mono_expo, _bounds_base, bounds_expo):
    # pylint: disable=too-many-return-statements, too-many-arguments
    if base < 0:
        return Monotonicity.Unknown
    elif almosteq(base, 0):
        return Monotonicity.Constant
    elif 0 < base < 1:
        if mono_expo.is_nondecreasing() and bounds_expo.is_nonpositive():
            return Monotonicity.Nondecreasing
        elif mono_expo.is_nonincreasing() and bounds_expo.is_nonnegative():
            return Monotonicity.Nondecreasing
        return Monotonicity.Unknown
    elif almostgte(base, 1):
        if mono_expo.is_nondecreasing() and bounds_expo.is_nonnegative():
            return Monotonicity.Nondecreasing
        elif mono_expo.is_nonincreasing() and bounds_expo.is_nonpositive():
            return Monotonicity.Nondecreasing
        return Monotonicity.Unknown
    return Monotonicity.Unknown # pragma: no cover


def _monotonicity_constant_exponent(base, expo, mono_base, mono_expo, bounds_base, bounds_expo):
    # pylint: disable=too-many-return-statements, too-many-arguments
    if almosteq(expo, 1):
        return mono_base
    elif almosteq(expo, 0):
        return Monotonicity.Constant
    is_integer = almosteq(expo, int(expo))
    is_even = almosteq(expo % 2, 0)
    if is_integer and is_even:
        return _monotonicity_even_exponent(
            base, expo, mono_base, mono_expo, bounds_base, bounds_expo,
        )
    elif is_integer:  # is odd
        return _monotonicity_odd_exponent(
            base, expo, mono_base, mono_expo, bounds_base, bounds_expo,
        )
    return _monotonicity_noninteger_exponent(
        base, expo, mono_base, mono_expo, bounds_base, bounds_expo,
    )


def _monotonicity_even_exponent(base, expo, mono_base, mono_expo, bounds_base, bounds_expo):
    # pylint: disable=too-many-return-statements, too-many-arguments, no-else-return
    if expo > 0:
        return _monotonicity_even_positive_exponent(
            base, expo, mono_base, mono_expo, bounds_base, bounds_expo,
        )
    else:
        return _monotonicity_even_negative_exponent(
            base, expo, mono_base, mono_expo, bounds_base, bounds_expo,
        )


def _monotonicity_even_positive_exponent(_base, _expo, mono_base, _mono_expo,
                                         bounds_base, _bounds_expo):
    # pylint: disable=too-many-return-statements, too-many-arguments
    if mono_base.is_nondecreasing() and bounds_base.is_nonnegative():
        return Monotonicity.Nondecreasing
    elif mono_base.is_nonincreasing() and bounds_base.is_nonpositive():
        return Monotonicity.Nondecreasing
    elif mono_base.is_nondecreasing() and bounds_base.is_nonpositive():
        return Monotonicity.Nonincreasing
    elif mono_base.is_nonincreasing() and bounds_base.is_nonnegative():
        return Monotonicity.Nonincreasing
    return Monotonicity.Unknown # pragma: no cover


def _monotonicity_even_negative_exponent(_base, _expo, mono_base, _mono_expo,
                                         bounds_base, _bounds_expo):
    # pylint: disable=too-many-return-statements, too-many-arguments
    if mono_base.is_nonincreasing() and bounds_base.is_nonnegative():
        return Monotonicity.Nondecreasing
    elif mono_base.is_nondecreasing() and bounds_base.is_nonpositive():
        return Monotonicity.Nondecreasing
    elif mono_base.is_nonincreasing() and bounds_base.is_nonpositive():
        return Monotonicity.Nonincreasing
    elif mono_base.is_nondecreasing() and bounds_base.is_nonnegative():
        return Monotonicity.Nonincreasing
    return Monotonicity.Unknown # pragma: no cover


def _monotonicity_odd_exponent(_base, expo, mono_base, _mono_expo, _bounds_base, _bounds_expo):
    # pylint: disable=too-many-return-statements, too-many-arguments
    if expo > 0 and mono_base.is_nondecreasing():
        return Monotonicity.Nondecreasing
    elif expo < 0 and mono_base.is_nonincreasing():
        return Monotonicity.Nondecreasing
    elif expo > 0 and mono_base.is_nonincreasing():
        return Monotonicity.Nonincreasing
    elif expo < 0 and mono_base.is_nondecreasing():
        return Monotonicity.Nonincreasing
    return Monotonicity.Unknown # pragma: no cover


def _monotonicity_noninteger_exponent(_base, expo, mono_base, _mono_expo,
                                      bounds_base, _bounds_expo):
    # pylint: disable=too-many-return-statements, too-many-arguments
    if not bounds_base.is_nonnegative():
        return Monotonicity.Unknown
    elif expo > 0:
        return mono_base
    elif expo < 0:
        if mono_base.is_nondecreasing():
            return Monotonicity.Nonincreasing
        elif mono_base.is_nonincreasing():
            return Monotonicity.Nondecreasing
    return Monotonicity.Unknown # pragma: no cover



# pylint: skip-file
import pytest
from pyomo.core.kernel.component_map import ComponentMap
from tests.conftest import PlaceholderExpression
from suspect.interval import Interval
from suspect.pyomo.expressions import UnaryFunctionExpression
from suspect.fbbt.initialization import BoundsInitializationVisitor
from suspect.fbbt.initialization.rules import (
    SqrtRule,
    LogRule,
    AsinRule,
    AcosRule,
)


def _visit(func_name):
    p = PlaceholderExpression()
    visitor = BoundsInitializationVisitor()
    root = UnaryFunctionExpression(p, name=func_name)
    bounds = ComponentMap()
    assert visitor.visit(root, bounds)
    print(bounds._dict)
    return bounds[p]


def test_sqrt():
    assert _visit('sqrt').is_nonnegative()


def test_log():
    assert _visit('log').is_nonnegative()


def test_asin():
    assert _visit('asin') == Interval(-1, 1)


def test_acos():
    assert _visit('acos') == Interval(-1, 1)



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Polynomial degree class."""


class PolynomialDegree(object):
    """Represent a polynomial degree."""
    def __init__(self, degree):
        self.degree = degree

    @classmethod
    def not_polynomial(cls):
        """Constructor for non polynomial polynomial."""
        return PolynomialDegree(None)

    def is_polynomial(self):
        """Predicate to check if it's polynomial."""
        return self.degree is not None

    def is_constant(self):
        """Predicate to check if it's constant."""
        return self.is_polynomial() and self.degree == 0

    def is_linear(self):
        """Predicate to check if it's linear."""
        return self.is_polynomial() and self.degree == 1

    def is_quadratic(self):
        """Predicate to check if it's quadratic."""
        return self.is_polynomial() and self.degree == 2

    def __add__(self, other):
        if self.is_polynomial() and other.is_polynomial():
            return PolynomialDegree(self.degree + other.degree)
        return self.not_polynomial()

    def __pow__(self, other):
        if isinstance(other, (int, float)):
            return PolynomialDegree(self.degree * other)
        return PolynomialDegree.not_polynomial()

    def __gt__(self, other):
        if not self.is_polynomial():
            return True
        if not other.is_polynomial():
            return False
        return self.degree > other.degree

    def __eq__(self, other):
        return self.degree == other.degree

    def __str__(self):
        return 'PolynomialDegree(degree={})'.format(self.degree)

    def __repr__(self):
        return '<{} at {}>'.format(str(self), hex(id(self)))



# pylint: skip-file
import pytest
from hypothesis import assume, given, reproduce_failure
import hypothesis.strategies as st
import itertools
import numpy as np
from pyomo.core.kernel.component_map import ComponentMap
import pyomo.environ as pe
from tests.strategies import coefficients, reals, variables, constants, expressions
from tests.conftest import PlaceholderExpression as PE, BilinearTerm
from suspect.expression import ExpressionType as ET, UnaryFunctionType as UFT
from suspect.monotonicity import Monotonicity as M
from suspect.convexity import Convexity as C
from suspect.convexity.visitor import ConvexityPropagationVisitor
from suspect.interval import Interval as I
from suspect.pyomo.expressions import *
from suspect.convexity.rules import *
from suspect.math import almosteq, pi


@pytest.fixture
def visitor():
    return ConvexityPropagationVisitor()


@st.composite
def convex_terms(draw):
    coef = draw(coefficients())
    if coef > 0:
        return (coef, C.Convex)
    else:
        return (coef, C.Concave)


@st.composite
def concave_terms(draw):
    coef = draw(coefficients())
    if coef < 0:
        return (coef, C.Convex)
    else:
        return (coef, C.Concave)


@st.composite
def linear_terms(draw):
    coef = draw(coefficients())
    return (coef, C.Linear)


@given(variables())
def test_variable_is_linear(visitor, var):
    matched, result = visitor.visit_expression(var, None, None, None)
    assert matched
    assert result.is_linear()


@given(constants())
def test_constant_is_linear(visitor, const):
    matched, result = visitor.visit_expression(const, None, None, None)
    assert matched
    assert result.is_linear()


@pytest.mark.parametrize('cvx_child,bounded_below,bounded_above,expected', [
    # l <= g(x) <= u is always unknown except if g(x) is linear
    (C.Linear, True, True, C.Linear),
    (C.Convex, True, True, C.Unknown),
    (C.Concave, True, True, C.Unknown),
    (C.Unknown, True, True, C.Unknown),

    # g(x) <= u is the same as g(x)
    (C.Linear, False, True, C.Linear),
    (C.Convex, False, True, C.Convex),
    (C.Concave, False, True, C.Concave),
    (C.Unknown, False, True, C.Unknown),

    # l <= g(x) is the negation of g(x)
    (C.Linear, True, False, C.Linear),
    (C.Convex, True, False, C.Concave),
    (C.Concave, True, False, C.Convex),
    (C.Unknown, True, False, C.Unknown),
])
@given(child=expressions())
def test_constraint(visitor, child, cvx_child, bounded_below, bounded_above, expected):
    if bounded_below:
        lower_bound = 0.0
    else:
        lower_bound = None

    if bounded_above:
        upper_bound = 1.0
    else:
        upper_bound = None

    expr = Constraint('constr', lower_bound, upper_bound, children=[child])

    convexity = ComponentMap()
    convexity[child] = cvx_child

    matched, result = visitor.visit_expression(expr, convexity, None, None)
    assert matched
    assert result == expected


@pytest.mark.parametrize('cvx_child,expected', [
    (C.Linear, C.Linear), (C.Convex, C.Convex), (C.Concave, C.Concave), (C.Unknown, C.Unknown),
])
@given(child=expressions())
def test_objective_is_minimizing(visitor, child, cvx_child, expected):
    expr = Objective('obj', children=[child])
    convexity = ComponentMap()
    convexity[child] = cvx_child

    matched, result = visitor.visit_expression(expr, convexity, None, None)
    assert matched
    assert result == expected


@pytest.mark.parametrize('cvx_child,expected', [
    (C.Linear, C.Linear), (C.Convex, C.Concave),
    (C.Concave, C.Convex), (C.Unknown, C.Unknown),
])
@given(child=expressions())
def test_objective_is_maximizing(visitor, child, cvx_child, expected):
    expr = Objective('obj', children=[child], sense=Sense.MAXIMIZE)
    convexity = ComponentMap()
    convexity[child] = cvx_child

    matched, result = visitor.visit_expression(expr, convexity, None, None)
    assert matched
    assert result == expected


@pytest.mark.parametrize('cvx,expected', [
    (C.Convex, C.Concave), (C.Concave, C.Convex),
    (C.Linear, C.Linear), (C.Unknown, C.Unknown),
])
@given(child=expressions())
def test_negation(visitor, child, cvx, expected):
    convexity = ComponentMap()
    convexity[child] = cvx
    expr = -child
    assume(isinstance(expr, NegationExpression))
    matched, result = visitor.visit_expression(expr, convexity, None, None)
    assert matched
    assert result == expected


class TestProduct:
    def _rule_result(self, visitor, f, g, cvx_f, cvx_g, mono_f, mono_g, bounds_f, bounds_g):
        convexity = ComponentMap()
        convexity[f] = cvx_f
        convexity[g] = cvx_g
        mono = ComponentMap()
        mono[f] = mono_f
        mono[g] = mono_g
        bounds = ComponentMap()
        bounds[f] = bounds_f
        bounds[g] = bounds_g
        expr = f * g
        assume(isinstance(expr, ProductExpression))
        matched, result = visitor.visit_expression(expr, convexity, mono, bounds)
        assert matched
        return result

    @pytest.mark.parametrize('cvx_f,bounds_g,expected', [
        (C.Convex, I(0, None), C.Convex),
        (C.Convex, I(None, 0), C.Concave),
        (C.Concave, I(0, None), C.Concave),
        (C.Concave, I(None, 0), C.Convex),
        (C.Linear, I(0, None), C.Linear),
        (C.Linear, I(None, 0), C.Linear),
        (C.Unknown, I(0, None), C.Unknown),
        (C.Unknown, I(None, 0), C.Unknown),
    ])
    @given(f=expressions(), g=constants())
    def test_product_with_constant(self, visitor, f, g, cvx_f, bounds_g, expected):
        assume(f.is_expression_type() and not isinstance(f, MonomialTermExpression))
        cvx_g = C.Linear # g is constant
        mono_f = M.Unknown
        mono_g = M.Constant
        bounds_f = I(None, None)
        assert self._rule_result(visitor, f, g, cvx_f, cvx_g, mono_f, mono_g, bounds_f, bounds_g) == expected
        assert self._rule_result(visitor, g, f, cvx_g, cvx_f, mono_g, mono_f, bounds_g, bounds_f) == expected

    @pytest.mark.parametrize('cvx_f,bounds_f,expected', [
        (C.Linear, I(None, None), C.Convex),
        (C.Linear, I(0, None), C.Convex),
        (C.Linear, I(None, 0), C.Convex),

        (C.Convex, I(None, None), C.Unknown),
        (C.Convex, I(0, None), C.Convex),
        (C.Convex, I(None, 0), C.Unknown),

        (C.Concave, I(None, None), C.Unknown),
        (C.Concave, I(0, None), C.Unknown),
        (C.Concave, I(None, 0), C.Convex),
    ])
    @given(f=expressions())
    def test_product_with_itself(self, visitor, f, cvx_f, bounds_f, expected):
        convexity = ComponentMap()
        convexity[f] = cvx_f
        bounds = ComponentMap()
        bounds[f] = bounds_f
        expr = f * f
        matched, result = visitor.visit_expression(expr, convexity, None, bounds)
        assert matched
        assert result == expected

    @given(var=variables(), coef=reals())
    def test_product_with_itself_with_coeff(self, visitor, var, coef):
        if coef > 0:
            expected = C.Convex
        else:
            expected = C.Concave
        rule = ProductRule()
        g = coef * var
        assume(isinstance(g, ProductExpression))
        matched, result = visitor.visit_expression(ProductExpression([var, g]), None, None, None)
        assert result == expected
        matched, result = visitor.visit_expression(ProductExpression([g, var]), None, None, None)
        assert result == expected

    @given(
        var=variables(),
        vars_with_coef=st.lists(
            st.tuples(
                variables(),
                constants(),
            ),
        )
    )
    def test_product_linear_by_var(self, visitor, var, vars_with_coef):
        rule = ProductRule()
        mono = ComponentMap()
        lin = sum(v * c for v, c in vars_with_coef)
        mono[var] = M.Nondecreasing
        mono[lin] = M.Nondecreasing
        matched, result = visitor.visit_expression(ProductExpression([var, lin]), None, mono, None)
        assert result == C.Unknown
        matched, result = visitor.visit_expression(ProductExpression([lin, var]), None, mono, None)
        assert result == C.Unknown


@pytest.mark.skip('Pyomo has no division')
class TestDivision:
    def _rule_result(self, cvx_f, cvx_g, mono_f, mono_g, bounds_f, bounds_g):
        rule = DivisionRule()
        f = PE()
        g = PE()
        convexity = ComponentMap()
        convexity[f] = cvx_f
        convexity[g] = cvx_g
        mono = ComponentMap()
        mono[f] = mono_f
        mono[g] = mono_g
        bounds = ComponentMap()
        bounds[f] = bounds_f
        bounds[g] = bounds_g
        return rule.apply(PE(ET.Division, [f, g]), convexity, mono, bounds)

    @pytest.mark.parametrize('cvx_f,bounds_g,expected', [
        (C.Convex, I(1e-5, None), C.Convex),
        (C.Convex, I(0, None), C.Unknown),

        (C.Concave, I(None, -1e-5), C.Convex),
        (C.Concave, I(None, 0), C.Unknown),

        (C.Convex, I(None, -1e-5), C.Concave),
        (C.Convex, I(None, 0), C.Unknown),

        (C.Concave, I(1e-5, None), C.Concave),
        (C.Concave, I(0, None), C.Unknown),
    ])
    def test_constant_denominator(self, cvx_f, bounds_g, expected):
        assert self._rule_result(cvx_f, C.Linear, M.Unknown, M.Constant, I(None, None), bounds_g) == expected


    @pytest.mark.parametrize('cvx_g,bounds_f,bounds_g,expected', [
        (C.Concave, I(0, None), I(1e-5, None), C.Convex),
        (C.Concave, I(0, None), I(0, None), C.Unknown),
        (C.Convex, I(None, 0), I(None, -1e-5), C.Convex),
        (C.Convex, I(None, 0), I(None, 0), C.Unknown),
        (C.Convex, I(0, None), I(None, -1e-5), C.Concave),
        (C.Convex, I(0, None), I(None, 0), C.Unknown),
        (C.Concave, I(None, 0), I(1e-5, None), C.Concave),
        (C.Concave, I(None, 0), I(0, None), C.Unknown),
    ])
    def test_constant_numerator(self, cvx_g, bounds_f, bounds_g, expected):
        assert self._rule_result(C.Linear, cvx_g, M.Constant, M.Unknown, bounds_f, bounds_g) == expected


@pytest.mark.skip('Pyomo has no linear expression')
class TestLinear:
    def _rule_result(self, terms):
        coefficients = [c for c, _ in terms]
        children = [PE() for _ in terms]
        cvxs = [c for _, c in terms]
        ctx = ConvexityContext()
        for child, cvx in zip(children, cvxs):
            ctx.set_convexity(child, cvx)
        rule = LinearRule()
        return rule.checked_apply(PE(ET.Linear, children=children, coefficients=coefficients), ctx)

    @given(
        st.lists(convex_terms(), min_size=1),
        st.lists(linear_terms()),
    )
    def test_convex(self, a, b):
        cvx = self._rule_result(a + b)
        assert cvx.is_convex()

    @given(
        st.lists(concave_terms(), min_size=1),
        st.lists(linear_terms()),
    )
    def test_concave(self, a, b):
        cvx = self._rule_result(a + b)
        assert cvx.is_concave()

    @given(
        st.lists(linear_terms(), min_size=1),
    )
    def test_linear(self, a):
        cvx = self._rule_result(a)
        assert cvx.is_linear()

    @given(
        st.lists(convex_terms(), min_size=1),
        st.lists(concave_terms(), min_size=1),
    )
    def test_unknown(self, a, b):
        terms = a + b
        for c, _ in terms:
            assume(c != 0.0)
        cvx = self._rule_result(terms)
        assert cvx.is_unknown()


class TestSum:
    def _rule_result(self, visitor, children_with_cvx):
        children = [c for c, _ in children_with_cvx]

        convexity = ComponentMap()
        for child, cvx in children_with_cvx:
            convexity[child] = cvx
        expr = SumExpression(children)
        matched, result = visitor.visit_expression(expr, convexity, None, None)
        assert matched
        return result

    @given(
        st.lists(
            st.tuples(expressions(), st.just(C.Convex)),
            min_size=1
        ),
        st.lists(
            st.tuples(expressions(), st.just(C.Linear)),
        ),
    )
    def test_convex(self, visitor, a, b):
        cvx = self._rule_result(visitor, a + b)
        assert cvx.is_convex()

    @given(
        st.lists(
            st.tuples(expressions(), st.just(C.Concave)),
            min_size=1
        ),
        st.lists(
            st.tuples(expressions(), st.just(C.Linear))
        ),
    )
    def test_concave(self, visitor, a, b):
        cvx = self._rule_result(visitor, a + b)
        assert cvx.is_concave()

    @given(
        st.lists(
            st.tuples(expressions(), st.just(C.Linear)),
            min_size=1
        ),
    )
    def test_linear(self, visitor, a):
        cvx = self._rule_result(visitor, a)
        assert cvx.is_linear()

    @given(
        st.lists(
            st.tuples(expressions(), st.just(C.Concave)),
            min_size=1
        ),
        st.lists(
            st.tuples(expressions(), st.just(C.Convex)),
            min_size=1
        ),
    )
    def test_unknown(self, visitor, a, b):
        cvx = self._rule_result(visitor, a + b)
        assert cvx.is_unknown()


class UnaryFunctionTest:
    def _rule_result(self, visitor, g, cvx, mono_g, bounds_g, func):
        convexity = ComponentMap()
        convexity[g] = cvx
        mono = ComponentMap()
        mono[g] = mono_g
        bounds = ComponentMap()
        bounds[g] = bounds_g

        expr = func(g)
        matched, result = visitor.visit_expression(expr, convexity, mono, bounds)
        assert matched
        return result


class TestAbs(UnaryFunctionTest):
    @pytest.mark.parametrize('mono,bounds',
                             itertools.product([M.Nondecreasing, M.Nonincreasing, M.Constant],
                                               [I(None, 0), I(0, None), I(None, None)]))
    @given(child=expressions())
    def test_linear_child(self, visitor, child, mono, bounds):
        cvx = self._rule_result(visitor, child, C.Linear, mono, bounds, abs)
        assert cvx == C.Convex

    @pytest.mark.parametrize('mono,bounds,expected', [
        (M.Unknown, I(0, None), C.Convex),
        (M.Unknown, I(None, 0), C.Concave),
    ])
    @given(child=expressions())
    def test_convex_child(self, visitor, child, mono, bounds, expected):
        cvx = self._rule_result(visitor, child, C.Convex, mono, bounds, abs)
        assert cvx == expected

    @pytest.mark.parametrize('mono,bounds,expected', [
        (M.Unknown, I(0, None), C.Concave),
        (M.Unknown, I(None, 0), C.Convex),
    ])
    @given(child=expressions())
    def test_concave_child(self, visitor, child, mono, bounds, expected):
        cvx = self._rule_result(visitor, child, C.Concave, mono, bounds, abs)
        assert cvx == expected




class TestSqrt(UnaryFunctionTest):
    @pytest.mark.parametrize(
        'cvx,mono',
        itertools.product(
            [C.Concave, C.Linear],
            [M.Nondecreasing, M.Nonincreasing, M.Constant, M.Unknown],
        )
    )
    @given(child=expressions())
    def test_concave_child(self, visitor, child, cvx, mono):
        cvx = self._rule_result(visitor, child, cvx, mono, I(0, None), pe.sqrt)
        assert cvx == C.Concave

    @pytest.mark.parametrize(
        'cvx,mono',
        itertools.product(
            [C.Convex, C.Unknown],
            [M.Nondecreasing, M.Nonincreasing, M.Constant, M.Unknown],
        ),
    )
    @given(child=expressions())
    def test_non_concave_child(self, visitor, child, cvx, mono):
        cvx = self._rule_result(visitor, child, cvx, mono, I(0, None), pe.sqrt)
        assert cvx == C.Unknown


class TestExp(UnaryFunctionTest):
    @pytest.mark.parametrize(
        'cvx,mono,bounds',
        itertools.product(
            [C.Convex, C.Linear],
            [M.Nondecreasing, M.Nonincreasing, M.Constant, M.Unknown],
            [I(0, None), I(None, 0), I(None, None)],
        )
    )
    @given(child=expressions())
    def test_convex_child(self, visitor, child, cvx, mono, bounds):
        cvx = self._rule_result(visitor, child, cvx, mono, bounds, pe.exp)
        assert cvx == C.Convex

    @pytest.mark.parametrize(
        'cvx,mono,bounds',
        itertools.product(
            [C.Concave, C.Unknown],
            [M.Nondecreasing, M.Nonincreasing, M.Constant, M.Unknown],
            [I(0, None), I(None, 0), I(None, None)],
        )
    )
    @given(child=expressions())
    def test_convex_child(self, visitor, child, cvx, mono, bounds):
        cvx = self._rule_result(visitor, child, cvx, mono, bounds, pe.exp)
        assert cvx == C.Unknown


class TestLog(UnaryFunctionTest):
    @pytest.mark.parametrize(
        'cvx,mono',
        itertools.product(
            [C.Concave, C.Linear],
            [M.Nondecreasing, M.Nonincreasing, M.Constant, M.Unknown],
        )
    )
    @given(child=expressions())
    def test_concave_child(self, visitor, child, cvx, mono):
        cvx = self._rule_result(visitor, child, cvx, mono, I(0, None), pe.log)
        assert cvx == C.Concave

    @pytest.mark.parametrize(
        'cvx,mono',
        itertools.product(
            [C.Convex, C.Unknown],
            [M.Nondecreasing, M.Nonincreasing, M.Constant, M.Unknown],
        ),
    )
    @given(child=expressions())
    def test_non_concave_child(self, visitor, child, cvx, mono):
        cvx = self._rule_result(visitor, child, cvx, mono, I(0, None), pe.log)
        assert cvx == C.Unknown


class TestSin(UnaryFunctionTest):
    @given(expressions())
    def test_bound_size_too_big(self, visitor, child):
        cvx = self._rule_result(visitor, child, C.Linear, M.Constant, I(-2, 2), pe.sin)
        assert cvx == C.Unknown

    @given(expressions())
    def test_bound_opposite_sign(self, visitor, child):
        cvx = self._rule_result(visitor, child, C.Linear, M.Constant, I(-0.1, 0.1), pe.sin)
        assert cvx == C.Unknown

    @given(expressions())
    def test_positive_sin_linear_child(self, visitor, child):
        cvx = self._rule_result(visitor, child, C.Linear, M.Constant, I(pi/2-0.1, pi/2+0.1), pe.sin)
        assert cvx == C.Concave

    @given(expressions())
    def test_positive_sin_convex_child_but_wrong_interval(self, visitor, child):
        cvx = self._rule_result(visitor, child, C.Convex, M.Unknown, I(pi/2-0.1, pi/2+0.1), pe.sin)
        assert cvx == C.Unknown

    @given(expressions())
    def test_positive_sin_convex_child(self, visitor, child):
        cvx = self._rule_result(visitor, child, C.Convex, M.Unknown, I(pi/2+0.1, pi-0.1), pe.sin)
        assert cvx == C.Concave

    @given(expressions())
    def test_positive_sin_concave_child(self, visitor, child):
        cvx = self._rule_result(visitor, child, C.Concave, M.Unknown, I(0.1, pi/2-0.1), pe.sin)
        assert cvx == C.Concave

    @given(expressions())
    def test_negative_sin_linear_child(self, visitor, child):
        cvx = self._rule_result(visitor, child, C.Linear, M.Constant, I(1.5*pi-0.1, 1.5*pi+0.1), pe.sin)
        assert cvx == C.Convex

    @given(expressions())
    def test_negative_sin_concave_child_but_wrong_interval(self, visitor, child):
        cvx = self._rule_result(visitor, child, C.Concave, M.Unknown, I(1.5*pi-0.1, 1.5*pi+0.1), pe.sin)
        assert cvx == C.Unknown

    @given(expressions())
    def test_negative_sin_concave_child(self, visitor, child):
        cvx = self._rule_result(visitor, child, C.Concave, M.Constant, I(pi+0.1, 1.5*pi-0.1), pe.sin)
        assert cvx == C.Convex

    @given(expressions())
    def test_negative_sin_convex_child(self, visitor, child):
        cvx = self._rule_result(visitor, child, C.Convex, M.Unknown, I(1.5*pi, 2*pi), pe.sin)
        assert cvx == C.Convex


class TestCos(UnaryFunctionTest):
    @given(expressions())
    def test_bound_size_too_big(self, visitor, child):
        cvx = self._rule_result(visitor, child, C.Linear, M.Constant, I(-2, 2), pe.cos)
        assert cvx == C.Unknown

    @given(expressions())
    def test_bound_opposite_sign(self, visitor, child):
        cvx = self._rule_result(visitor, child, C.Linear, M.Constant, I(pi/2-0.1, pi/2+0.1), pe.cos)
        assert cvx == C.Unknown

    @given(expressions())
    def test_positive_cos_linear_child(self, visitor, child):
        cvx = self._rule_result(visitor, child, C.Linear, M.Constant, I(0, 0.5*pi), pe.cos)
        assert cvx == C.Concave

    @given(expressions())
    def test_positive_cos_convex_child_but_wrong_interval(self, visitor, child):
        cvx = self._rule_result(visitor, child, C.Convex, M.Unknown, I(0, 0.5*pi), pe.cos)
        assert cvx == C.Unknown

    @given(expressions())
    def test_positive_cos_convex_child(self, visitor, child):
        cvx = self._rule_result(visitor, child, C.Convex, M.Unknown, I(1.5*pi, 2*pi), pe.cos)
        assert cvx == C.Concave

    @given(expressions())
    def test_positive_cos_concave_child(self, visitor, child):
        cvx = self._rule_result(visitor, child, C.Concave, M.Unknown, I(0, 0.5*pi), pe.cos)
        assert cvx == C.Concave

    @given(expressions())
    def test_negative_cos_linear_child(self, visitor, child):
        cvx = self._rule_result(visitor, child, C.Linear, M.Constant, I(0.6*pi, 1.4*pi), pe.cos)
        assert cvx == C.Convex

    @given(expressions())
    def test_negative_cos_concave_child_but_wrong_interval(self, visitor, child):
        cvx = self._rule_result(visitor, child, C.Concave, M.Unknown, I(0.6*pi, 0.9*pi), pe.cos)
        assert cvx == C.Unknown

    @given(expressions())
    def test_negative_cos_concave_child(self, visitor, child):
        cvx = self._rule_result(visitor, child, C.Concave, M.Unknown, I(1.1*pi, 1.4*pi), pe.cos)
        assert cvx == C.Convex

    @given(expressions())
    def test_negative_cos_convex_child(self, visitor, child):
        cvx = self._rule_result(visitor, child, C.Convex, M.Unknown, I(0.6*pi, 0.9*pi), pe.cos)
        assert cvx == C.Convex



class TestTan(UnaryFunctionTest):
    @given(expressions())
    def test_bound_size_too_big(self, visitor, child):
        cvx = self._rule_result(visitor, child, C.Linear, M.Constant, I(-2, 2), pe.tan)
        assert cvx == C.Unknown

    @given(expressions())
    def test_bound_opposite_sign(self, visitor, child):
        cvx = self._rule_result(visitor, child, C.Linear, M.Constant, I(-0.1, 0.1), pe.tan)
        assert cvx == C.Unknown

    @given(expressions())
    def test_positive_tan_convex_child(self, visitor, child):
        cvx = self._rule_result(visitor, child, C.Convex, M.Unknown, I(pi, 1.5*pi), pe.tan)
        assert cvx == C.Convex

    @given(expressions())
    def test_positive_tan_concave_child(self, visitor, child):
        cvx = self._rule_result(visitor, child, C.Concave, M.Unknown, I(pi, 1.5*pi), pe.tan)
        assert cvx == C.Unknown

    @given(expressions())
    def test_negative_tan_convex_child(self, visitor, child):
        cvx = self._rule_result(visitor, child, C.Convex, M.Unknown, I(-1.5*pi, -pi), pe.tan)
        assert cvx == C.Unknown

    @given(expressions())
    def test_negative_tan_concave_child(self, visitor, child):
        cvx = self._rule_result(visitor, child, C.Concave, M.Unknown, I(-0.49*pi, -0.1), pe.tan)
        assert cvx == C.Concave


class TestAsin(UnaryFunctionTest):
    @given(expressions())
    def test_is_concave(self, visitor, child):
        cvx = self._rule_result(visitor, child, C.Concave, M.Unknown, I(-1.0, 0), pe.asin)
        assert cvx == C.Concave

    @pytest.mark.parametrize('bounds', [I(-1, 0.1), I(-1.1, 0.0), I(None, None)])
    @given(child=expressions())
    def test_is_not_concave(self, visitor, child, bounds):
        cvx = self._rule_result(visitor, child, C.Concave, M.Unknown, bounds, pe.asin)
        assert cvx == C.Unknown

    @given(child=expressions())
    def test_is_convex(self, visitor, child):
        cvx = self._rule_result(visitor, child, C.Convex, M.Unknown, I(0, 1), pe.asin)
        assert cvx == C.Convex

    @pytest.mark.parametrize('bounds', [I(-0.1, 1), I(0.0, 1.1), I(None, None)])
    @given(child=expressions())
    def test_is_not_convex(self, visitor, child, bounds):
        cvx = self._rule_result(visitor, child, C.Convex, M.Unknown, bounds, pe.asin)
        assert cvx == C.Unknown


class TestAcos(UnaryFunctionTest):
    @given(child=expressions())
    def test_is_concave(self, visitor, child):
        cvx = self._rule_result(visitor, child, C.Convex, M.Unknown, I(0, 1), pe.acos)
        assert cvx == C.Concave

    @pytest.mark.parametrize('bounds', [I(-0.1, 1), I(0.0, 1.1), I(None, None)])
    @given(child=expressions())
    def test_is_not_concave(self, visitor, child, bounds):
        cvx = self._rule_result(visitor, child, C.Convex, M.Unknown, bounds, pe.acos)
        assert cvx == C.Unknown

    @given(child=expressions())
    def test_is_convex(self, visitor, child):
        cvx = self._rule_result(visitor, child, C.Concave, M.Unknown, I(-1, 0), pe.acos)
        assert cvx == C.Convex

    @pytest.mark.parametrize('bounds', [I(-1, 0.1), I(-1.1, 0.0), I(None, None)])
    @given(child=expressions())
    def test_is_not_convex(self, visitor, child, bounds):
        cvx = self._rule_result(visitor, child, C.Concave, M.Unknown, bounds, pe.acos)
        assert cvx == C.Unknown


class TestAtan(UnaryFunctionTest):
    @given(child=expressions())
    def test_is_concave(self, visitor, child):
        cvx = self._rule_result(visitor, child, C.Concave, M.Unknown, I(0, None), pe.atan)
        assert cvx == C.Concave

    @pytest.mark.parametrize('bounds', [I(None, 0), I(None, None)])
    @given(child=expressions())
    def test_is_not_concave(self, visitor, child, bounds):
        cvx = self._rule_result(visitor, child, C.Concave, M.Unknown, bounds, pe.atan)
        assert cvx == C.Unknown

    @given(child=expressions())
    def test_is_convex(self, visitor, child):
        cvx = self._rule_result(visitor, child, C.Convex, M.Unknown, I(None, 0), pe.atan)
        assert cvx == C.Convex

    @pytest.mark.parametrize('bounds', [I(0, None), I(None, None)])
    @given(child=expressions())
    def test_is_not_convex(self, visitor, child, bounds):
        cvx = self._rule_result(visitor, child, C.Convex, C.Unknown, bounds, pe.atan)
        assert cvx == C.Unknown


class TestPowConstantBase:
    def _rule_result(self, visitor, base, expo, cvx_expo, mono_expo, bounds_expo):
        convexity = ComponentMap()
        convexity[expo] = cvx_expo
        convexity[base] = C.Linear
        mono = ComponentMap()
        mono[expo] = mono_expo
        mono[base] = M.Constant
        bounds = ComponentMap()
        bounds[base] = I(base, base)
        bounds[expo] = bounds_expo

        expr = PowExpression([base, expo])
        matched, result = visitor.visit_expression(expr, convexity, mono, bounds)
        assert matched
        return result

    @pytest.mark.parametrize(
        'cvx_expo,mono_expo,bounds_expo',
        itertools.product(
            [C.Convex, C.Concave, C.Linear, C.Unknown],
            [M.Nondecreasing, M.Nonincreasing, M.Unknown],
            [I(None, 0), I(0, None), I(None, None)],
        )
    )
    @given(
        base=reals(max_value=-0.01, allow_infinity=False),
        expo=expressions(),
    )
    def test_negative_base(self, visitor, base, expo, cvx_expo, mono_expo, bounds_expo):
        cvx = self._rule_result(visitor, base, expo, cvx_expo, mono_expo, bounds_expo)
        assert cvx == C.Unknown

    @pytest.mark.parametrize('cvx_expo,mono_expo,bounds_expo',
        itertools.product(
            [C.Convex, C.Concave, C.Linear, C.Unknown],
            [M.Nondecreasing, M.Nonincreasing, M.Unknown],
            [I(None, 0), I(0, None), I(None, None)],
        )
    )
    @given(
        base=reals(min_value=0.001, max_value=0.999),
        expo=expressions(),
    )
    def test_base_between_0_and_1(self, visitor, base, expo, cvx_expo, mono_expo, bounds_expo):
        if cvx_expo == C.Concave or cvx_expo == C.Linear:
            expected = C.Convex
        else:
            expected = C.Unknown
        cvx = self._rule_result(visitor, base, expo, cvx_expo, mono_expo, bounds_expo)
        assert cvx == expected

    @pytest.mark.parametrize('cvx_expo,mono_expo,bounds_expo',
        itertools.product(
            [C.Convex, C.Concave, C.Linear, C.Unknown],
            [M.Nondecreasing, M.Nonincreasing, M.Unknown],
            [I(None, 0), I(0, None), I(None, None)],
        )
    )
    @given(
        base=reals(min_value=1, allow_infinity=False),
        expo=expressions(),
    )
    def test_base_gt_1(self, visitor, base, expo, cvx_expo, mono_expo, bounds_expo):
        if cvx_expo == C.Convex or cvx_expo == C.Linear:
            expected = C.Convex
        else:
            expected = C.Unknown
        cvx = self._rule_result(visitor, base, expo, cvx_expo, mono_expo, bounds_expo)
        assert cvx == expected


class TestPowConstantExponent(object):
    def _rule_result(self, visitor, base, cvx_base, mono_base, bounds_base, expo):
        convexity = ComponentMap()
        convexity[base] = cvx_base
        convexity[expo] = C.Linear
        mono = ComponentMap()
        mono[base] = mono_base
        mono[expo] = M.Constant
        bounds = ComponentMap()
        bounds[base] = bounds_base
        bounds[expo] = I(expo, expo)
        expr = PowExpression([base, expo])
        matched, result = visitor.visit_expression(expr, convexity, mono, bounds)
        assert matched
        return result

    @pytest.mark.parametrize(
        'cvx_base,mono_base,bounds_base',
        itertools.product(
            [C.Convex, C.Concave, C.Linear, C.Unknown],
            [M.Nondecreasing, M.Nonincreasing, M.Unknown],
            [I(None, 0), I(0, None), I(None, None)],
        )
    )
    @given(base=expressions())
    def test_exponent_equals_0(self, visitor, base, cvx_base, mono_base, bounds_base):
        cvx = self._rule_result(visitor, base, cvx_base, mono_base, bounds_base, 0.0)
        assert cvx == C.Linear

    @pytest.mark.parametrize(
        'cvx_base,mono_base,bounds_base',
        itertools.product(
            [C.Convex, C.Concave, C.Linear, C.Unknown],
            [M.Nondecreasing, M.Nonincreasing, M.Unknown],
            [I(None, 0), I(0, None), I(None, None)],
        )
    )
    @given(base=expressions())
    def test_exponent_equals_1(self, visitor, base, cvx_base, mono_base, bounds_base):
        cvx = self._rule_result(visitor, base, cvx_base, mono_base, bounds_base, 1.0)
        assert cvx == cvx_base

    @pytest.mark.parametrize('cvx_base,mono_base,bounds_base,expected', [
        (C.Linear, M.Nondecreasing, I(None, None), C.Convex),

        (C.Convex, M.Unknown, I(0, None), C.Convex),
        (C.Convex, M.Unknown, I(None, 0), C.Unknown),

        (C.Concave, M.Unknown, I(0, None), C.Unknown),
        (C.Concave, M.Unknown, I(None, 0), C.Convex),
    ])
    @given(
        base=expressions(),
        expo=st.integers(min_value=1, max_value=1000),
    )
    def test_positive_even_integer(self, visitor, base, expo, cvx_base, mono_base,
                                   bounds_base, expected):
        cvx = self._rule_result(visitor, base, cvx_base, mono_base, bounds_base, 2*expo)
        assert cvx == expected

    @pytest.mark.parametrize('cvx_base,mono_base,bounds_base,expected', [
        (C.Convex, M.Unknown, I(None, 0), C.Convex),
        (C.Convex, M.Unknown, I(0, None), C.Concave),

        (C.Concave, M.Unknown, I(0, None), C.Convex),
        (C.Concave, M.Unknown, I(None, 0), C.Concave),
    ])
    @given(
        base=expressions(),
        expo=st.integers(min_value=1, max_value=1000),
    )
    def test_negative_even_integer(self, visitor, base, expo, cvx_base, mono_base,
                                   bounds_base, expected):
        cvx = self._rule_result(visitor, base, cvx_base, mono_base, bounds_base, -2*expo)
        assert cvx == expected

    @pytest.mark.parametrize('cvx_base,mono_base,bounds_base,expected', [
        (C.Convex, M.Unknown, I(0, None), C.Convex),
        (C.Convex, M.Unknown, I(None, 0), C.Unknown),
        (C.Concave, M.Unknown, I(None, 0), C.Concave),
        (C.Concave, M.Unknown, I(0, None), C.Unknown),
    ])
    @given(
        base=expressions(),
        expo=st.integers(min_value=1, max_value=1000),
    )
    def test_positive_odd_integer(self, visitor, base, expo, cvx_base, mono_base,
                                  bounds_base, expected):
        cvx = self._rule_result(visitor, base, cvx_base, mono_base, bounds_base, 2*expo+1)
        assert cvx == expected

    @pytest.mark.parametrize('cvx_base,mono_base,bounds_base,expected', [
        (C.Concave, M.Unknown, I(0, None), C.Convex),
        (C.Concave, M.Unknown, I(None, 0), C.Unknown),
        (C.Convex, M.Unknown, I(None, 0), C.Concave),
        (C.Convex, M.Unknown, I(0, None), C.Unknown),
    ])
    @given(
        base=expressions(),
        expo=st.integers(min_value=1, max_value=1000),
    )
    def test_negative_odd_integer(self, visitor, base, expo, cvx_base, mono_base,
                                  bounds_base, expected):
        cvx = self._rule_result(visitor, base, cvx_base, mono_base, bounds_base, -2*expo+1)
        assert cvx == expected

    @given(
        base=expressions(),
        expo=reals(min_value=1, allow_infinity=False, max_value=1000),
    )
    def test_positive_gt_1_non_integer_negative_base(self, visitor, base, expo):
        expo = expo + 1e-3
        assume(expo != int(expo))
        cvx = self._rule_result(visitor, base, C.Convex, M.Unknown, I(None, -1), expo)
        assert cvx == C.Unknown

    @given(
        base=expressions(),
        expo=reals(min_value=1, allow_infinity=False, max_value=1000),
    )
    def test_positive_gt_1_non_integer(self, visitor, base, expo):
        expo = expo + 1e-5 # make it positive
        assume(expo != int(expo))
        cvx = self._rule_result(visitor, base, C.Convex, M.Unknown, I(0, None), expo)
        assert cvx == C.Convex

    @pytest.mark.parametrize('cvx,expected', [(C.Convex, C.Concave), (C.Concave, C.Convex)])
    @given(
        base=expressions(),
        expo=reals(min_value=-1000, max_value=0, allow_infinity=False),
    )
    def test_positive_lt_0_non_integer(self, visitor, base, expo, cvx, expected):
        expo = expo - 1e-5 # make it negative
        assume(not almosteq(expo, int(expo)))
        cvx = self._rule_result(visitor, base, cvx, M.Unknown, I(0, None), expo)
        assert cvx == expected

    @given(
        base=expressions(),
        expo=reals(min_value=0, max_value=1, allow_infinity=False),
    )
    def test_positive_0_1_non_integer(self, visitor, base, expo):
        assume(not almosteq(expo, int(expo)))
        cvx = self._rule_result(visitor, base, C.Concave, M.Unknown, I(0, None), expo)
        assert cvx == C.Concave


@pytest.mark.skip('Pyomo has no quadratic')
class TestQuadratic:
    def _rule_result(self, A):
        n = A.shape[0]
        var = [PE(ET.Variable) for _ in range(n)]
        terms = []
        for i in range(n):
            for j in range(i+1):
                terms.append(BilinearTerm(var[i], var[j], A[i, j]))

        expr = PE(ET.Quadratic, terms=terms, children=var)
        rule = QuadraticRule()
        return rule.apply(expr, None)

    @given(coefs=st.lists(reals(min_value=0, allow_infinity=False), min_size=1))
    def test_sum_of_squares_is_convex_with_positive_coefficients(self, coefs):
        assume(any([c > 0 and not np.isclose(c, 0) for c in coefs]))
        A = np.eye(len(coefs)) * coefs
        cvx = self._rule_result(A)
        assert cvx == C.Convex

    @given(coefs=st.lists(reals(max_value=0, allow_infinity=False), min_size=1))
    def test_sum_of_squares_is_concave_with_negative_coefficients(self, coefs):
        assume(any([c < 0 and not np.isclose(c, 0) for c in coefs]))
        A = np.eye(len(coefs)) * coefs
        cvx = self._rule_result(A)
        assert cvx == C.Concave

    @given(
        neg_coefs=st.lists(reals(max_value=0.0, allow_infinity=False), min_size=1),
        pos_coefs=st.lists(reals(min_value=0.0, allow_infinity=False), min_size=1))
    def test_sum_of_squares_is_unknown_otherwise(self, neg_coefs, pos_coefs):
        assume(any([n < 0 and not np.isclose(n, 0) for n in neg_coefs]))
        assume(any([p > 0 and not np.isclose(p, 0) for p in pos_coefs]))
        coefs = neg_coefs + pos_coefs
        A = np.eye(len(coefs)) * coefs
        cvx = self._rule_result(A)
        assert cvx == C.Unknown

    @given(st.integers(min_value=1, max_value=100))
    def test_positive_definite_is_convex(self, n):
        B = np.random.randn(n, n)
        A = np.eye(n) * n + 0.5 * (B + B.T)
        cvx = self._rule_result(A)
        assert cvx == C.Convex

    @given(st.integers(min_value=1, max_value=100))
    def test_negative_definite_is_concave(self, n):
        B = np.random.randn(n, n)
        A = -np.eye(n) * n - 0.5 * (B + B.T)
        cvx = self._rule_result(A)
        assert cvx == C.Concave



# pylint: skip-file
import pytest
import pyomo.environ as aml
from suspect import detect_special_structure
from suspect.convexity import Convexity


@pytest.fixture
def problem():
    m = aml.ConcreteModel()
    m.x = aml.Var(bounds=(-2.0, 2.0))
    m.y = aml.Var(bounds=(0.01, 1.5))

    m.obj = aml.Objective(expr=m.x + m.y)
    m.c0 = aml.Constraint(expr=m.x**2 + m.y ** 2 <= 1.0)
    m.c1 = aml.Constraint(expr=aml.sin(m.y) >= 0.0)

    return m


def test_summary(problem):
    info = detect_special_structure(problem)
    assert info.num_variables() == 2
    assert info.num_integers() == 0
    assert info.num_binaries() == 0
    assert info.num_constraints() == 2
    conscurvature =  info.conscurvature()
    assert conscurvature['c0'] == Convexity.Convex
    assert conscurvature['c1'] == Convexity.Convex

    objcurvature = info.objcurvature()
    assert objcurvature['obj'] == Convexity.Linear

    objtype = info.objtype()
    assert objtype['obj'] == 'linear'



# pylint: skip-file
import hypothesis.strategies as st
import pyomo.environ as pe
import pytest
from hypothesis import given, assume, settings, HealthCheck
from pyomo.core.kernel.component_map import ComponentMap

from suspect.polynomial.rules import *
from suspect.polynomial.visitor import PolynomialDegreeVisitor
from suspect.pyomo.expressions import (
    MonomialTermExpression,
    DivisionExpression,
    SumExpression,
    Constraint,
    Objective,
    PowExpression,
)
from tests.strategies import expressions, variables, reals, unary_function_types


@pytest.fixture
def visitor():
    return PolynomialDegreeVisitor()


@st.composite
def polynomial_degrees(draw, allow_not_polynomial=True):
    if allow_not_polynomial:
        degree = draw(st.one_of(st.none(), st.integers(min_value=0)))
    else:
        degree = draw(st.integers(min_value=0))
    return PolynomialDegree(degree)


def test_variable_is_always_1(visitor):
    matched, result = visitor.visit_expression(pe.Var(), None)
    assert matched
    assert result.degree == 1


def test_constant_is_always_0(visitor):
    matched, result = visitor.visit_expression(123.0, None)
    assert matched
    assert result.degree == 0


@given(expressions(), polynomial_degrees())
def test_objective(visitor, child, child_degree):
    expr = Objective('obj', children=[child])
    poly = ComponentMap()
    poly[child] = child_degree
    matched, result = visitor.visit_expression(expr, poly)
    assert matched
    assert result == child_degree


@given(expressions(), polynomial_degrees())
def test_constraint(visitor, child, child_degree):
    expr = Constraint('cons', 0.0, None, children=[child])
    poly = ComponentMap()
    poly[child] = child_degree
    matched, result = visitor.visit_expression(expr, poly)
    assert matched
    assert result == child_degree


@given(expressions(), polynomial_degrees())
def test_negation(visitor, child, child_degree):
    expr = -child
    assume(child.is_expression_type() and not isinstance(expr, MonomialTermExpression))
    poly = ComponentMap()
    poly[child] = child_degree
    matched, result = visitor.visit_expression(expr, poly)
    assert matched
    assert result == child_degree


def test_division_rule_with_constant_denominator(visitor):
    child = pe.Var()
    num = 1.0
    poly = ComponentMap()
    poly[num] = PolynomialDegree(0)
    poly[child] = PolynomialDegree(0)
    expr = DivisionExpression([num, child])
    matched, result = visitor.visit_expression(expr, poly)
    assert matched
    assert result.is_constant()


@given(polynomial_degrees())
def test_division_rule_with_nonconstant_denominator(visitor, den_degree):
    assume(not den_degree.is_constant())
    child = pe.Var()
    num = 1.0
    expr = DivisionExpression([num, child])
    poly = ComponentMap()
    poly[num] = PolynomialDegree(0)
    poly[child] = den_degree
    matched, result = visitor.visit_expression(expr, poly)
    assert matched
    assert not result.is_polynomial()



@given(expressions(), expressions(), polynomial_degrees(), polynomial_degrees())
def test_product_rule(visitor, f, g, poly_f, poly_g):
    expr = f * g
    poly = ComponentMap()
    poly[f] = poly_f
    poly[g] = poly_g
    matched, result = visitor.visit_expression(expr, poly)
    assert matched
    if poly_f.is_polynomial() and poly_g.is_polynomial():
        assert result.is_polynomial()
        assert result.degree == poly_f.degree + poly_g.degree
    else:
        assert not result.is_polynomial()


@pytest.mark.skip('Linear Expression not implemented.')
@given(st.integers(min_value=1, max_value=100))
def test_linear_rule_with_children(children_size):
    rule = LinearRule()
    children = [PE(ET.Variable)] * children_size
    result = rule.apply(PE(ET.Linear, children), None)
    assert result.degree == 1


@settings(suppress_health_check=[HealthCheck.too_slow])
@given(st.lists(st.tuples(expressions(), polynomial_degrees()), min_size=2, max_size=4))
def test_sum_rule(visitor, data):
    children = [c for c, _ in data]
    for c in children:
        assume(c.is_expression_type() and not isinstance(c, SumExpression))
    degrees = [d for _, d in data]
    poly = ComponentMap()
    for c, d in data:
        poly[c] = d
    expr = sum(children)
    print(expr, expr.args)
    matched, result = visitor.visit_expression(expr, poly)
    assert matched
    if all(d.is_polynomial() for d in degrees):
        assert result.is_polynomial()
        assert result.degree == max(d.degree for d in degrees)
    else:
        assert not result.is_polynomial()


@given(expressions(), expressions(), polynomial_degrees())
def test_power_with_non_polynomial_exponent(visitor, base, expo, base_poly):
    poly = ComponentMap()
    poly[base] = base_poly
    poly[expo] = PolynomialDegree(None)
    expr = base ** expo
    matched, result = visitor.visit_expression(expr, poly)
    assert matched
    assert not result.is_polynomial()


@given(expressions(), expressions(), polynomial_degrees())
def test_power_with_non_polynomial_base(visitor, base, expo, expo_poly):
    poly = ComponentMap()
    poly[base] = PolynomialDegree(None)
    poly[expo] = expo_poly
    expr = base ** expo
    matched, result = visitor.visit_expression(expr, poly)
    assert matched
    assert not result.is_polynomial()


@given(variables(), reals(min_value=1.1))
def test_power_constant_power_constant(visitor, v, c):
    expr = v ** c
    poly = ComponentMap()
    poly[v] = PolynomialDegree(0)
    poly[c] = PolynomialDegree(0)
    matched, result = visitor.visit_expression(expr, poly)
    assert matched
    assert result.is_polynomial() and result.degree == 0


@given(variables(), reals(allow_infinity=False))
def test_power_non_constant(visitor, v, c):
    assume(c != int(c))
    expr = v ** c
    poly = ComponentMap()
    poly[v] = PolynomialDegree(1)
    poly[c] = PolynomialDegree(0)
    matched, result = visitor.visit_expression(expr, poly)
    assert matched
    assert not result.is_polynomial()


@given(variables(), expressions())
def test_power_non_constant_polynomial(visitor, v, c):
    expr = v ** c
    poly = ComponentMap()
    poly[v] = PolynomialDegree(1)
    poly[c] = PolynomialDegree(4)
    matched, result = visitor.visit_expression(expr, poly)
    assert matched
    assert not result.is_polynomial()


@given(expressions(), st.integers(min_value=1), st.integers(min_value=2))
def test_power_constant_value(visitor, base, base_poly, expo):
    expr = base ** expo
    assume(isinstance(expr, PowExpression))
    poly = ComponentMap()
    poly[base] = PolynomialDegree(base_poly)
    poly[expo] = PolynomialDegree(0)
    matched, result = visitor.visit_expression(expr, poly)
    assert matched
    assert result.is_polynomial() and result.degree == base_poly * expo


@given(expressions(), polynomial_degrees(), unary_function_types)
def test_unary_function(visitor, child, poly_child, func_name):
    if func_name == 'abs':
        expr = abs(child)
        poly = ComponentMap()
        poly[child] = poly_child

        matched, result = visitor.visit_expression(expr, poly)
        assert matched
        assert result == poly_child

    else:
        expr = getattr(pe, func_name)(child)

        matched, result = visitor.visit_expression(expr, None)
        assert matched
        assert not result.is_polynomial()



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Convexity detection rules for quadratic expressions."""
import numpy as np

from suspect.math import almostgte, almostlte
from pyomo.environ import ComponentMap
from suspect.convexity.convexity import Convexity
from suspect.convexity.rules.rule import ConvexityRule


def _gershgorin_circles_test(expr, var_to_idx):
    """Check convexity by computing Gershgorin circles without building
    the coefficients matrix.

    If the circles lie in the nonnegative (nonpositive) space, then the matrix
    is positive (negative) definite.

    Parameters
    ----------
    expr : QuadraticExpression
        the quadratic expression
    var_to_idx : dict-like
        map a var to a numerical index between 0 and n, where n is the number
        of args of expr

    Returns
    -------
    Convexity if the expression is Convex or Concave, None otherwise.
    """
    n = len(var_to_idx)
    row_circles = np.zeros(n)
    diagonal = np.zeros(n)

    for term in expr.terms:
        i = var_to_idx[term.var1]
        j = var_to_idx[term.var2]
        if i == j:
            diagonal[i] = term.coefficient
        else:
            coef = np.abs(term.coefficient / 2.0)
            row_circles[j] += coef
            row_circles[i] += coef

    diagonal_minus_row_circles = diagonal - row_circles
    diagonal_plus_row_circles = diagonal + row_circles

    if np.all(almostgte(diagonal_minus_row_circles, 0.0)):
        return Convexity.Convex
    if np.all(almostlte(diagonal_plus_row_circles, 0.0)):
        return Convexity.Concave
    return None


def _eigval_test(A):
    """Check convexity by computing eigenvalues.

    Parameters
    ----------
    A : matrix
        the (symmetric) coefficients matrix

    Returns
    -------
    Convexity if the expression is Convex or Concave, None otherwise.
    """
    eigv = np.linalg.eigvalsh(A)
    if np.all(eigv >= 0):
        return Convexity.Convex
    elif np.all(eigv <= 0):
        return Convexity.Concave
    return None


def _build_var_to_idx_map(expr):
    map = ComponentMap()
    idx = 0
    for term in expr.terms:
        v1 = term.var1
        v2 = term.var2
        if v1 not in map:
            map[v1] = idx
            idx += 1
        if v2 not in map:
            map[v2] = idx
            idx += 1
    return map


class QuadraticRule(ConvexityRule):
    """Return convexity of quadratic."""
    def __init__(self, max_expr_children=None):
        if max_expr_children is None:
            max_expr_children = 100
        self.max_expr_children = max_expr_children

    def apply(self, expr, convexity, mono, bounds):
        var_to_idx = _build_var_to_idx_map(expr)
        convexity = _gershgorin_circles_test(expr, var_to_idx)

        if convexity is not None:
            return convexity

        # Build coefficient matrix
        n = len(var_to_idx)
        if self.max_expr_children and n > self.max_expr_children:
            return Convexity.Unknown

        A = np.zeros((n, n))
        for term in expr.terms:
            i = var_to_idx[term.var1]
            j = var_to_idx[term.var2]
            if term.var1 is term.var2:
                A[i, j] = term.coefficient
            else:
                A[i, j] = term.coefficient / 2.0
                A[j, i] = A[i, j]

        convexity = _eigval_test(A)
        if convexity is not None:
            return convexity

        return Convexity.Unknown



# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: skip-file
import pytest
import numpy as np
from collections import namedtuple
from suspect.interval import Interval
from suspect.monotonicity.monotonicity import Monotonicity
from suspect.convexity.convexity import Convexity
from suspect.expression import ExpressionType as ET
from suspect.context import SpecialStructurePropagationContext



BilinearTerm = namedtuple('BilinearTerm', ['var1', 'var2', 'coefficient'])


class PlaceholderExpression(object):
    depth = 0
    def __init__(self, expression_type=None, children=None, coefficients=None,
                 is_constant=False, value=None, constant_term=None, terms=None,
                 bounded_above=False, bounded_below=False,
                 lower_bound=None, upper_bound=None,
                 is_minimizing=False, func_type=None):
        self.expression_type = expression_type
        self.children = children
        if coefficients is not None:
            self._coefficients = dict([(ch, co) for ch, co in zip(children, coefficients)])
        else:
            self._coefficients = {}

        self.is_constant = lambda: is_constant
        self.constant_term = constant_term
        self.terms = terms
        self.value = value
        self.bounded_above = lambda: bounded_above
        self.bounded_below = lambda: bounded_below
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.is_minimizing = lambda: is_minimizing
        self.func_type = func_type

    def coefficient(self, expr):
        return self._coefficients[expr]

    def is_variable_type(self):
        return self.expression_type == ET.Variable

    def is_expression_type(self):
        return not (self.is_variable_type() or self.is_constant())

    @property
    def args(self):
        return self.children

    @property
    def lb(self):
        return self.lower_bound

    @property
    def ub(self):
        return self.upper_bound

@pytest.fixture
def ctx():
    return SpecialStructurePropagationContext()


def bound_description_to_bound(bound_str):
    if isinstance(bound_str, str):
        return {
            'zero': Interval.zero(),
            'nonpositive': Interval(None, 0),
            'nonnegative': Interval(0, None),
            'positive': Interval(1, None),
            'negative': Interval(None, -1),
            'unbounded': Interval(None, None),
        }[bound_str]
    elif isinstance(bound_str, Interval):
        return bound_str
    else:
        return Interval(bound_str, bound_str)


def mono_description_to_mono(mono_str):
    return {
        'nondecreasing': Monotonicity.Nondecreasing,
        'nonincreasing': Monotonicity.Nonincreasing,
        'constant': Monotonicity.Constant,
        'unknown': Monotonicity.Unknown,
    }[mono_str]


def cvx_description_to_cvx(cvx_str):
    return {
        'convex': Convexity.Convex,
        'concave': Convexity.Concave,
        'linear': Convexity.Linear,
        'unknown': Convexity.Unknown,
    }[cvx_str]



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Monotonicity detection rules for abs function."""
from suspect.monotonicity.monotonicity import Monotonicity
from suspect.monotonicity.rules.rule import MonotonicityRule
from suspect.expression import UnaryFunctionType


class AbsRule(MonotonicityRule):
    """Return monotonicity of abs."""
    def apply(self, expr, monotonicity, bounds):
        child = expr.args[0]
        mono = monotonicity[child]
        child_bounds = bounds[child]

        if mono.is_constant():
            return mono

        # good examples to understand the behaviour of abs are abs(-x) and
        # abs(1/x)
        if child_bounds.is_nonnegative():
            # abs(x), x > 0 is the same as x
            return mono
        elif child_bounds.is_nonpositive():
            # abs(x), x < 0 is the opposite of x
            return mono.negate()
        return Monotonicity.Unknown



# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import warnings

import pyomo.environ as pe

from suspect.fbbt import perform_fbbt
from suspect.interval import Interval
from suspect.convexity import Convexity
from suspect.polynomial_degree import polynomial_degree
from suspect.propagation import propagate_special_structure

logger = logging.getLogger('suspect')


class ModelInformation(object):
    """Hold structure information about a problem.

    Attributes
    ----------
    variables : dict
       map variables names to their information
    objectives : dict
       map objectives names to their information
    constraints : dict
       map constraints names to their information
    """
    def __init__(self, name, variables, objectives, constraints):
        self.name = name
        self.variables = variables
        self.objectives = objectives
        self.constraints = constraints

    def num_variables(self):
        """Returns the number of variables in the problem."""
        return len(self.variables)

    def num_binaries(self):
        """Returns the number of binary variables in the problem."""
        binaries = [
            v for v in self.variables.values()
            if v['type'] == 'binary'
        ]
        return len(binaries)

    def num_integers(self):
        """Returns the number of integer variables in the problem."""
        integers = [
            v for v in self.variables.values()
            if v['type'] == 'integer'
        ]
        return len(integers)

    def num_constraints(self):
        """Returns the number of constraints in the problem."""
        return len(self.constraints)

    def conscurvature(self):
        """Returns the convexity of each constraint."""
        return dict(
            (k, v['convexity'])
            for k, v in self.constraints.items()
        )

    def objcurvature(self):
        """Returns the convexity of each objective."""
        return dict(
            (k, v['convexity'])
            for k, v in self.objectives.items()
        )

    def objtype(self):
        """Returns the type of each objective."""
        def _objtype(v):
            cvx = v['convexity']
            deg = v['polynomial_degree']
            if cvx.is_linear():
                assert deg == 1
                return 'linear'
            elif deg == 2:
                return 'quadratic'
            elif deg is not None:
                return 'polynomial'
            else:
                return 'nonlinear'

        return dict(
            (k, _objtype(v))
            for k, v in self.objectives.items()
        )


def detect_special_structure(problem, max_iter=10):
    """Detect special structure in the problem.

    Parameters
    ----------
    model : suspect.dag.ProblemDag or ConcreteModel
        the problem DAG or a Pyomo ConcreteModel.

    max_iter : int
        number of maximum bound propagation/tightening iterations.

    Returns
    -------
    ModelInformation
        an object containing the detected infomation about the problem.
    """
    bounds = perform_fbbt(problem)

    monotonicity, convexity = propagate_special_structure(problem, bounds)

    variables = {}
    for variable in problem.component_data_objects(pe.Var, active=True, descend_into=True):
        if variable.is_binary():
            variable_type = 'binary'
        elif variable.is_integer():
            variable_type = 'integer'
        else:
            variable_type = 'continuous'
        var_bounds = bounds[variable]
        variable_name = variable.name

        variables[variable_name] = {
            'name': variable_name,
            'type': variable_type,
            'lower_bound': var_bounds.lower_bound,
            'upper_bound': var_bounds.upper_bound,
        }

    objectives = {}
    for obj in problem.component_data_objects(pe.Objective, active=True, descend_into=True):
        if obj.is_minimizing():
            sense = 'min'
        else:
            sense = 'max'
        obj_bounds = bounds.get(obj.expr, Interval(None, None))
        cvx = convexity[obj.expr]
        poly = obj.expr.polynomial_degree()

        if sense == 'max':
            cvx = cvx.negate()

        objectives[obj.name] = {
            'sense': sense,
            'convexity': cvx,
            'polynomial_degree': poly,
            'lower_bound': obj_bounds.lower_bound,
            'upper_bound': obj_bounds.upper_bound,
        }

    constraints = {}
    for cons in problem.component_data_objects(pe.Constraint, active=True, descend_into=True):
        if cons.has_lb() and cons.has_ub():
            type_ = 'equality'
        else:
            type_ = 'inequality'

        cons_bounds = bounds.get(cons.body, Interval(None, None))
        cvx = convexity[cons.body]
        poly = cons.body.polynomial_degree()

        if type_ == 'equality':
            if not cvx.is_linear():
                cvx = Convexity.Unknown
        else:
            if cons.has_lb():
                cvx = cvx.negate()

        constraints[cons.name] = {
            'type': type_,
            'convexity': cvx,
            'polynomial_degree': poly,
            'lower_bound': cons_bounds.lower_bound,
            'upper_bound': cons_bounds.upper_bound,
        }

    return ModelInformation(
        name=problem.name,
        variables=variables,
        objectives=objectives,
        constraints=constraints,
    )



# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
from suspect.math import (
    make_number,
    almosteq,
    almostlte,
)


class FloatHasher(abc.ABC):
    @abc.abstractmethod
    def hash(self, f):
        raise NotImplementedError('hash')


class BTreeFloatHasher(FloatHasher):
    """A floating point hasher that keeps all seen floating
    point numbers ina binary tree.

    Good if the unique values of the floating point numbers in
    the problem are relatively few.
    """

    class Node(object):
        def __init__(self, num, hash_, left=None, right=None):
            self.num = num
            self.hash = hash_
            self.left = left
            self.right = right

    def __init__(self):
        self.root = None
        self.node_count = 0

    def hash(self, f):
        f = make_number(f)
        if self.root is None:
            self.root = self._make_node(f)
            return self.root.hash

        curr_node = self.root
        while True:
            if almosteq(f, curr_node.num):
                return curr_node.hash
            elif almostlte(f, curr_node.num):
                if curr_node.left is None:
                    new_node = self._make_node(f)
                    curr_node.left = new_node
                    return new_node.hash
                else:
                    curr_node = curr_node.left
            else:
                if curr_node.right is None:
                    new_node = self._make_node(f)
                    curr_node.right = new_node
                    return new_node.hash
                else:
                    curr_node = curr_node.right

    def _make_node(self, f):
        node = self.Node(f, self.node_count, None, None)
        self.node_count += 1
        return node


class RoundFloatHasher(FloatHasher):
    """A float hasher that hashes floats up to the n-th
    decimal place.
    """
    def __init__(self, n=2):
        self.n = 10**n

    def hash(self, f):
        return hash(int(f * self.n))



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""FBBT bounds initialization rules."""
from suspect.interval import Interval
from suspect.interfaces import UnaryFunctionRule


class _UnaryFunctionBoundsRule(UnaryFunctionRule):
    initial_bound = None

    def apply(self, expr, _bounds):
        child = expr.args[0]
        return [self.initial_bound]


class SqrtRule(_UnaryFunctionBoundsRule):
    """Bound propagation rule for sqrt."""
    initial_bound = Interval(0, None)


class LogRule(_UnaryFunctionBoundsRule):
    """Bound propagation rule for log."""
    initial_bound = Interval(0, None)


class AsinRule(_UnaryFunctionBoundsRule):
    """Bound propagation rule for asin."""
    initial_bound = Interval(-1, 1)


class AcosRule(_UnaryFunctionBoundsRule):
    """Bound propagation rule for acos."""
    initial_bound = Interval(-1, 1)



import pytest
import requests
import pyomo.environ as pe
import contextlib
import tempfile
from suspect.pyomo import read_qplib, create_connected_model


def _download_qplib_file(file, problem_name):
    url = 'http://qplib.zib.de/qplib/QPLIB_{}.qplib'.format(problem_name)
    response = requests.get(url, stream=True)
    file.write(response.content)


@contextlib.contextmanager
def qplib_problem(problem_name):
    with tempfile.NamedTemporaryFile(suffix='.qplib') as f:
        _download_qplib_file(f, problem_name)
        f.seek(0)
        yield f


def read_qplib_problem(filename):
    result = read_qplib(filename)
    assert isinstance(result, pe.ConcreteModel)


@pytest.mark.parametrize('problem_name', ['0018', '0681', '0685', '3554'])
def test_read_qplib_problem(benchmark, problem_name):
    with qplib_problem(problem_name) as f:
        benchmark(read_qplib_problem, f.name)


@pytest.mark.parametrize('problem_name', ['0018', '0681', '0685', '3554'])
def test_convert_qplib_problem(benchmark, problem_name):
    with qplib_problem(problem_name) as f:
        model = read_qplib(f.name)
        connected_model, _ = benchmark(create_connected_model, model)
        assert isinstance(connected_model, pe.ConcreteModel)


def test_profile():
    with qplib_problem('3554') as f:
        model = read_qplib(f.name)
        connected_model, _ = create_connected_model(model)
        assert isinstance(connected_model, pe.ConcreteModel)



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""SUSPECT extra convexity detectors."""
from .quadratic import QuadraticFormConvexityDetector
from .fractional import FractionalConvexityDetector
from .norm import L2NormConvexityDetector
from .perspective import PerspectiveFunctionConvexityDetector


__all__ = [
    'QuadraticFormConvexityDetector', 'FractionalConvexityDetector',
    'L2NormConvexityDetector', 'PerspectiveFunctionConvexityDetector',
]



# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum


class Convexity(Enum):
    Convex = 0
    Concave = 1
    Linear = 2
    Unknown = 3

    def is_convex(self):
        return self == self.Convex or self == self.Linear

    def is_concave(self):
        return self == self.Concave or self == self.Linear

    def is_linear(self):
        return self == self.Linear

    def is_unknown(self):
        return self == self.Unknown

    def negate(self):
        if self.is_linear():
            return self.Linear
        elif self.is_convex():
            return self.Concave
        elif self.is_concave():
            return self.Convex
        else:
            return self.Unknown

    def combine(self, other):
        """Combines two Convexity objects together.

        If both objects are the same type (linear, convex, concave) then
        the resulting object will be of the same type, otherwise it will
        be Convexity.Unknown
        """
        if self.is_linear() and other.is_linear():
            return self.Linear
        elif self.is_convex() and other.is_convex():
            return self.Convex
        elif self.is_concave() and other.is_concave():
            return self.Concave
        else:
            return self.Unknown



# Copyright 2019 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tags are a way to tag a group of nodes as special structure."""
from pyomo.core.kernel.component_map import ComponentMap
import pyomo.environ as pe
from suspect.pyomo.expressions import nonpyomo_leaf_types
import suspect.pyomo.expressions as dex


class Tag(object):
    """A Tag represents a special structure expression that spans
    across multiple nodes.
    """

    @staticmethod
    def from_expr(problem, expr):
        """Create `Tag` from `expr`.

        Parameters
        ----------
        problem : dag
            the problem containing the expression
        expr : expression
            the root expression

        Returns
        -------
        Tag or None
            returns a new Tag if expression matches, otherwise None
        """
        raise NotImplementedError('Tag.from_expr')


class LinearExpressionTag(Tag):
    def __init__(self, problem, expr, coef_map, constant):
        self.problem = problem
        self.expr = expr
        self._coef_map = coef_map
        self.constant = constant

    @staticmethod
    def from_expr(problem, expr):
        if not isinstance(expr, dex.SumExpression):
            return None

        if not all(LinearExpressionTag._is_linear_type(a)
                   for a in expr.args):
            return None

        coef_map = ComponentMap()
        constant = 0.0
        for arg in expr.args:
            if LinearExpressionTag._is_var(arg):
                if arg not in coef_map:
                    coef_map[arg] = 0.0
                coef_map[arg] += 1.0
            elif LinearExpressionTag._is_monomial(arg):
                assert len(arg.args) == 2
                const, var = arg.args
                if var not in coef_map:
                    coef_map[var] = 0.0
                coef_map[var] += pe.value(const)
            else:
                assert LinearExpressionTag._is_constant(arg)
                constant += pe.value(arg)
        return LinearExpressionTag(problem, expr, coef_map, constant)

    def coefficient(self, var):
        return self._coef_map[var]

    @staticmethod
    def _is_linear_type(expr):
        return (
            LinearExpressionTag._is_monomial(expr) or
            LinearExpressionTag._is_var(expr) or
            LinearExpressionTag._is_constant(expr)
        )

    @staticmethod
    def _is_monomial(expr):
        return isinstance(expr, dex.MonomialTermExpression)

    @staticmethod
    def _is_var(expr):
        return expr.is_variable_type()

    @staticmethod
    def _is_constant(expr):
        return expr.is_constant()



class QuadraticExpressionTag(Tag):
    def __init__(self, problem, expr, coef_map):
        self.problem = problem
        self.expr = expr
        self._coef_map = coef_map

    @staticmethod
    def from_expr(problem, expr):
        if not isinstance(expr, dex.SumExpression):
            return None

        if not all(a.polynomial_degree() == 2 for a in expr.args):
            return None

        # build bilinear terms as sparse matrix
        return None

    @staticmethod
    def _vars_and_coef(self, expr):
        pass

    # cases:
    # x ** 2
    # xy
    # (3x)y
    # (xy)3
    # x(3y)
    def terms(self):
        pass



# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from numbers import Number
import pyomo.environ as aml
from pyomo.core.expr.numeric_expr import (
    nonpyomo_leaf_types,
    NumericConstant,
)
from pyomo.core.expr.logical_expr import (
    InequalityExpression,
    EqualityExpression,
    RangedExpression,
)
from suspect.interval import Interval


def is_numeric(expr):
    return (
        expr.__class__ in nonpyomo_leaf_types or
        isinstance(expr, NumericConstant)
    )


def numeric_value(expr):
    if expr.__class__ in nonpyomo_leaf_types:
        return float(expr)
    assert isinstance(expr, NumericConstant)
    return float(expr.value)


def bounds_and_expr(expr):
    if isinstance(expr, (InequalityExpression, RangedExpression)):
        return _inequality_bounds_and_expr(expr)
    elif isinstance(expr, EqualityExpression):
        return _equality_bounds_and_expr(expr)
    else:
        raise ValueError('expr must be InequalityExpression or EqualityExpression')


def _inequality_bounds_and_expr(expr):
    if len(expr._args_) == 2:
        (lhs, rhs) = expr._args_
        if is_numeric(lhs):
            return Interval(numeric_value(lhs), None), rhs
        else:
            return Interval(None, numeric_value(rhs)), lhs
    elif len(expr._args_) == 3:
        (lhs, ex, rhs) = expr._args_
        return Interval(numeric_value(lhs), numeric_value(rhs)), ex
    else:
        raise ValueError('Malformed InequalityExpression')


def _equality_bounds_and_expr(expr):
    if len(expr._args_) == 2:
        body, rhs = expr._args_
        return Interval(numeric_value(rhs), numeric_value(rhs)), body
    else:
        raise ValueError('Malformed EqualityExpression')


def model_variables(model):
    """Return a list of variables in the model"""
    for v in model.component_data_objects(aml.Var, active=True, descend_into=True, sort=True):
        yield v


def model_constraints(model):
    """Return a list of constraints in the model"""
    for c in model.component_data_objects(aml.Constraint, active=True, descend_into=True, sort=True):
        yield c


def model_objectives(model):
    """Return a list of objectives in the model"""
    for obj in model.component_data_objects(aml.Objective, active=True, descend_into=True, sort=True):
        yield obj



# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Class to run bound tightener or a problem."""
import pyomo.environ as pe

from pyomo.core.expr.visitor import StreamBasedExpressionVisitor
from pyomo.core.kernel.component_map import ComponentMap

from suspect.interval import Interval
from suspect.fbbt.initialization import initialize_bounds
from suspect.fbbt.propagation import propagate_bounds_leaf_to_root
from suspect.fbbt.tightening import tighten_bounds_root_to_leaf


def perform_fbbt(model, max_iter=10, active=True, objective_bounds=None, initial_bounds=None, should_continue=None):
    """Perform FBBT on the model.

    Parameters
    ----------
    model : pyomo.environ.ConcreteModel
        the pyomo model
    max_iter : int
        maximum number of FBBT iterations
    active : boolean
        perform FBBT only on active constraints
    objective_bounds : dict-like
        map between an objective and its lower and upper bound
    initial_bounds : dict-like
        map with existing bounds
    should_continue : callable
        predicate that returns False if fbbt should stop early

    Returns
    -------
    ComponentMap
        tightened bounds for variables and expressions
    """
    if max_iter < 1:
        raise ValueError("max_iter must be >= 1")

    if initial_bounds is None:
        bounds = ComponentMap()
    else:
        bounds = initial_bounds

    if should_continue is None:
        should_continue = lambda: True

    if objective_bounds is None:
        objective_bounds = ComponentMap()

    objectives = list(model.component_data_objects(pe.Objective, active=active, descend_into=True))
    constraints = list(model.component_data_objects(pe.Constraint, active=active, descend_into=True))

    for constraint in constraints:
        lower = pe.value(constraint.lower)
        upper = pe.value(constraint.upper)
        bounds[constraint.body] = Interval(lower, upper)

    for objective in objectives:
        obj_lower, obj_upper = objective_bounds.get(objective, (None, None))
        bounds[objective.expr] = Interval(obj_lower, obj_upper)

    # Force one full iteration of bounds propagation
    _perform_fbbt_step(model, bounds, -1, active, lambda: True, constraints, objectives)

    for iter in range(max_iter):
        changed = _perform_fbbt_step(model, bounds, iter, active, should_continue, constraints, objectives)
        if not changed:
            break
        if not should_continue():
            break

    return bounds


def _perform_fbbt_step(model, bounds, iter, active, should_continue, constraints, objectives):
    if iter < 0:
        func = _initialize_then_propagate_bounds_on_expr
    else:
        func = _tighten_then_propagate_bounds_on_expr

    changed = False
    for objective in objectives:
        changed |= func(objective.expr, bounds)
        if not should_continue():
            return changed

    for constraint in constraints:
        changed |= func(constraint.body, bounds)
        if not should_continue():
            return changed

    return changed


def _initialize_then_propagate_bounds_on_expr(expr, bounds):
    return _perform_fbbt_iteration_on_expr(
        expr, bounds, initialize_bounds, propagate_bounds_leaf_to_root, visit_all=True
    )


def _tighten_then_propagate_bounds_on_expr(expr, bounds):
    return _perform_fbbt_iteration_on_expr(
        expr, bounds, tighten_bounds_root_to_leaf, propagate_bounds_leaf_to_root
    )


def _perform_fbbt_iteration_on_expr(expr, bounds, tighten, propagate, visit_all=False):
    def enter_node(node):
        result = tighten(node, bounds)
        if result is None:
            return None, None
        if isinstance(result, list):
            result_iter = zip(expr.args, result)
        else:
            result_iter = result.items()
        descend_into_args = []
        for arg, value in result_iter:
            changed = _update_bounds_map(bounds, arg, value)
            if changed:
                descend_into_args.append(arg)

        if visit_all:
            return None, None

        return descend_into_args, None

    def exit_node(node, data):
        result = propagate(node, bounds)
        return _update_bounds_map(bounds, node, result)

    return StreamBasedExpressionVisitor(
        enterNode=enter_node,
        exitNode=exit_node,
    ).walk_expression(expr)


def _update_bounds_map(bounds, expr, value):
    if value is None:
        new_bounds = Interval(None, None)
    else:
        new_bounds = value

    old_bounds = bounds.get(expr, None)

    if old_bounds is not None:
        new_bounds = old_bounds.intersect(
            new_bounds,
            abs_eps=1e-6,
        )
        has_changed = old_bounds != new_bounds
    else:
        has_changed = True
    bounds[expr] = new_bounds
    return has_changed



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Convexity detector for quadratic expressions."""
import numpy as np
from suspect.ext import ConvexityDetector
from suspect.convexity import Convexity
from suspect.interfaces import Rule
from suspect.expression import ExpressionType


class QuadraticFormConvexityDetector(ConvexityDetector):
    """Detect convexity of quadratic expressions xQx."""
    def __init__(self, _problem):
        super().__init__()

    def visit_expression(self, expr, convexity, mono, bounds):
        return False, None
        return {
            ExpressionType.Product: QuadraticRule(),
        }


class QuadraticRule(Rule):
    """Detect convexity of quadratic expressions xQx."""
    root_expr = ExpressionType.Sum

    @property
    def _allowed_children(self):
        return (
            ExpressionType.Product,
            ExpressionType.Negation,
            ExpressionType.Power,
        )

    def apply(self, expr, ctx):
        if not ctx.polynomial(expr).is_quadratic():
            return

        # Check convexity of non quadratic children
        for child in expr.children:
            if not ctx.polynomial(child).is_quadratic():
                if not ctx.convexity(child).is_convex():
                    # we have sum of quadratic + unknown nonlinear
                    # It can't be convex so just stop checking
                    return
            else:
                if child.expression_type not in self._allowed_children:
                    return

        count = 0
        unique_vars = {}
        for (vi, vj, _) in self._variables_of_quadratic(expr, ctx):
            if vi not in unique_vars:
                unique_vars[vi] = count
                count += 1
            if vj not in unique_vars:
                unique_vars[vj] = count
                count += 1

        n = len(unique_vars)
        A = np.zeros((n, n))
        for (vi, vj, c) in self._variables_of_quadratic(expr, ctx):
            i = unique_vars[vi]
            j = unique_vars[vj]
            if i == j:
                A[i, j] = c
            else:
                A[i, j] = c / 2
                A[j, i] = A[i, j]
        eigv = np.linalg.eigvalsh(A)
        if np.all(eigv >= 0):
            return Convexity.Convex
        elif np.all(eigv <= 0):
            return Convexity.Concave

    def _variables_of_quadratic(self, expr, ctx):
        def _both_variables(a, b):
            return (a, b, 1.0)

        def _variable_and_linear(a, b):
            bc = b.children[0]
            return (a, bc, b.coefficient(bc))

        def _const_pow_expression(const, powe):
            a, b = powe.children
            a_type = a.expression_type
            b_type = b.expression_type

            if a_type == ExpressionType.Variable:
                assert b_type == ExpressionType.Constant
                return (a, a, const)
            elif b_type == ExpressionType.Variable:
                assert a_type == ExpressionType.Constant
                return (b, b, const)
            else:
                raise RuntimeError('unreachable')

        def _product_expression(expr):
            a, b = expr.children
            a_type = a.expression_type
            b_type = b.expression_type
            if a_type == ExpressionType.Variable and b_type == ExpressionType.Variable:
                return _both_variables(a, b)
            elif a_type == ExpressionType.Variable and b_type == ExpressionType.Linear:
                return _variable_and_linear(a, b)
            elif b_type == ExpressionType.Variable and a_type == ExpressionType.Linear:
                return _variable_and_linear(b, a)
            elif a_type == ExpressionType.Constant and b_type == ExpressionType.Power:
                return _const_pow_expression(a.value, b)
            elif b_type == ExpressionType.Constant and a_type == ExpressionType.Power:
                return _const_pow_expression(b.value, a)
            else:
                raise RuntimeError('unreachable')

        def _child_variables(child):
            child_type = child.expression_type
            if child_type == ExpressionType.Product:
                return _product_expression(child)
            elif child_type == ExpressionType.Power:
                return _const_pow_expression(1.0, child)
            elif child_type == ExpressionType.Negation:
                expr = child.children[0]
                return _child_variables(expr)
            else:
                raise RuntimeError('unreachable')

        for child in expr.children:
            if not ctx.polynomial[child].is_quadratic():
                continue
            yield _child_variables(child)



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""FBBT bounds initialization visitor."""
from suspect.interval import Interval
from suspect.interfaces import CombineUnaryFunctionRules
from suspect.pyomo.expressions import UnaryFunctionExpression, Constraint
from suspect.visitor import BackwardVisitor
import pyomo.environ as pyo
from suspect.fbbt.initialization.rules import (
    SqrtRule,
    LogRule,
    AsinRule,
    AcosRule,
)

_rule = CombineUnaryFunctionRules({
    'sqrt': SqrtRule(),
    'log': LogRule(),
    'asin': AsinRule(),
    'acos': AcosRule()},
    needs_matching_rules=False,
)


def initialize_bounds(expr, bounds):
    if isinstance(expr, pyo.Constraint):
        lower = pyo.value(expr.lower)
        upper = pyo.value(expr.upper)
        return [Interval(lower, upper)]
    if isinstance(expr, UnaryFunctionExpression):
        return _rule.apply(expr, bounds)
    return None


class BoundsInitializationVisitor(BackwardVisitor):
    """Initialize problem bounds using function domains as bound."""
    needs_matching_rules = False

    def visit_expression(self, expr, bounds):
        if isinstance(expr, UnaryFunctionExpression):
            return True, _rule.apply(expr, bounds)
        return False, None

    def handle_result(self, expr, value, bounds):
        if value is None:
            new_bounds = Interval(None, None)
        else:
            new_bounds = value
        bounds[expr] = new_bounds
        return True



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Interfaces used in SUSPECT.

You don't need them, but they are nice for documentation purpose.
"""
import abc
from suspect.expression import ExpressionType


class Problem(object): # pragma: no cover
    """Generic problem with vertices of type V."""
    pass


class Visitor(metaclass=abc.ABCMeta): # pragma: no cover
    """Visitor for vertices of Problem."""
    @abc.abstractmethod
    def visit(self, vertex, ctx):
        """Visit vertex. Return True if the vertex should be considered "dirty"."""
        pass


class Iterator(metaclass=abc.ABCMeta): # pragma: no cover
    """Iterator over vertices of Problem."""
    @abc.abstractmethod
    def iterate(self, problem, visitor, ctx, *args, **kwargs):
        """Iterate over vertices of problem, calling visitor on each one of them.

        Returns the list of vertices for which the visitor returned a True value.
        """
        pass


class ForwardIterator(Iterator): # pragma: no cover pylint: disable=abstract-method
    """An iterator for iterating over nodes in ascending depth order."""
    pass


class BackwardIterator(Iterator): # pragma: no cover pylint: disable=abstract-method
    """An iterator for iterating over nodes in descending depth order."""
    pass


class Rule(object): # pragma: no cover
    """Represent a series of contraints on an expression yielding a value."""
    def apply(self, expr, ctx):
        """Apply rule to ``expr`` and ``ctx``."""
        pass


class UnaryFunctionRule(Rule): # pragma: no cover
    """Represent a series of contraints on an unary function expression yielding a value."""
    pass


class CombineUnaryFunctionRules(Rule): # pragma: no cover
    """Rule to combine a collection of UnaryFunctionRule.

    Parameters
    ----------
    rules: UnaryFunctionRule dict
       dict of unary functions rules
    needs_matching_rules : bool
       if True, will raise an exception if no rule matched
    """
    def __init__(self, rules, needs_matching_rules=True):
        self._rules = rules
        self._needs_matching_rules = needs_matching_rules
        self._apply_funcs = {}
        for func_name, rule in self._rules.items():
            self._apply_funcs[func_name] = rule.apply

    def apply(self, expr, ctx, *args):
        apply_func = self._apply_funcs.get(expr.getname())
        if not apply_func:
            if not self._needs_matching_rules:
                return None
            if hasattr(expr, getname):
                func_name = expr.getname()
            else:
                func_name = 'Unknown'
            raise RuntimeError(
                'Could not find rule for expression of expression_type={} and func_name={}'.format(
                    getattr(expr, 'expression_type'),
                    func_name,
                )
            )
        return apply_func(expr, ctx, *args)



# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Numerical interval object."""

from functools import wraps
# pylint: disable=no-name-in-module
from suspect.math import (
    almosteq,
    almostgte,
    almostlte,
    inf,
    zero,
    pi,
    down,
    up,
    make_number,
    min_,
    max_,
    exp,
    log,
    power,
    sqrt,
    sin,
    cos,
    tan,
    asin,
    acos,
    atan,
)
from suspect.math import RoundMode as RM


class EmptyIntervalError(ValueError):
    pass


class _FunctionOnInterval(object):
    def __init__(self, func, inv_func):
        self._func = func
        self._inv_func = inv_func

    def __call__(self):
        return self._func()

    @property
    def inverse(self):
        """Return the inverse of this function."""
        return _FunctionOnInterval(self._inv_func, self._func)


def monotonic_increasing(func):
    """Correctly round monotonic increasing function."""
    # pylint: disable=invalid-name, protected-access
    @wraps(func)
    def _wrapper(it):
        inner = func(it)
        return Interval(inner(it._lower, RM.RD), inner(it._upper, RM.RU))
    return _wrapper


def monotonic_decreasing(func):
    """Correctly round monotonic decreasing function."""
    # pylint: disable=invalid-name, protected-access
    @wraps(func)
    def _wrapper(it):
        inner = func(it)
        return Interval(inner(it._upper, RM.RD), inner(it._lower, RM.RU))
    return _wrapper


class Interval: # pylint: disable=too-many-public-methods
    """Numerical interval."""
    __slots__ = ('_lower', '_upper', '_is_zero')

    def __init__(self, lower, upper):
        if lower is None:
            lower = -inf
        if upper is None:
            upper = inf

        lower = make_number(lower)
        upper = make_number(upper)

        if lower > upper:
            raise ValueError('lower must be <= upper')

        self._lower = lower
        self._upper = upper
        self._is_zero = None

    @staticmethod
    def zero():
        """Return Interval = [0, 0]."""
        return Interval(zero, zero)

    @property
    def lower_bound(self):
        """Return interval lower bound as float."""
        return float(self._lower)

    @property
    def upper_bound(self):
        """Return interval upper bound as float."""
        return float(self._upper)

    def size(self):
        """Return interval size."""
        if self._lower == -inf or self._upper == inf:
            return inf
        return self._upper - self._lower

    def is_zero(self):
        """Check if the interval is [0, 0]."""
        if self._is_zero is None:
            self._is_zero = \
                almosteq(self._lower, 0) and almosteq(self._upper, 0)
        return self._is_zero

    def is_positive(self):
        """Check if the interval [a, b] has a > 0."""
        return self._lower > zero

    def is_negative(self):
        """Check if the interval [a, b] has b < 0."""
        return self._upper < zero

    def is_nonpositive(self):
        """Check if the interval [a, b] has b <= 0."""
        return almostlte(self._upper, zero)

    def is_nonnegative(self):
        """Check if the interval [a, b] has a >= 0."""
        return almostgte(self._lower, zero)

    def inverse(self):
        """Return the inverse of self, or 1/self."""
        return Interval(
            down(lambda: 1.0 / self._upper),
            up(lambda: 1.0 / self._lower),
        )

    def intersect(self, other, rel_eps=None, abs_eps=None):
        """Intersect this interval with another."""
        if not isinstance(other, Interval):
            raise ValueError('intersect with non Interval value')

        new_lower = max(self._lower, other._lower)  # pylint: disable=protected-access
        new_upper = min(self._upper, other._upper)  # pylint: disable=protected-access

        if new_upper < new_lower:
            if almosteq(new_lower, new_upper, rel_eps=rel_eps, abs_eps=abs_eps):
                new_lower = new_upper
            else:
                raise EmptyIntervalError()
        return Interval(new_lower, new_upper)

    @property
    def negation(self):
        """Return the negation of self, or 0-self"""
        return _FunctionOnInterval(lambda: -self, lambda: -self)

    @property
    def sqrt(self):  # pragma: no cover
        """Return the bound of sqrt(self)"""
        return _FunctionOnInterval(self._sqrt, self._sqr)

    @property
    def exp(self):  # pragma: no cover
        """Return the bound of exp(self)"""
        return _FunctionOnInterval(self._exp, self._log)

    @property
    def log(self):  # pragma: no cover
        """Return the bound of log(self)"""
        return _FunctionOnInterval(self._log, self._exp)

    @property
    def sin(self):
        """Return the bound of sin(self)"""
        return _FunctionOnInterval(self._sin, self._asin)

    @property
    def cos(self):
        """Return the bound of cos(self)"""
        return _FunctionOnInterval(self._cos, self._acos)

    @property
    def tan(self):
        """Return the bound of tan(self)"""
        return _FunctionOnInterval(self._tan, self._atan)

    @property
    def asin(self):
        """Return the bound of asin(self)"""
        return _FunctionOnInterval(self._asin, self._sin)

    @property
    def acos(self):
        """Return the bound of acos(self)"""
        return _FunctionOnInterval(self._acos, self._cos)

    @property
    def atan(self):
        """Return the bound of atan(self)"""
        return _FunctionOnInterval(self._atan, self._tan)

    def __pos__(self):
        return self

    def __neg__(self):
        return Interval(-self._upper, -self._lower)

    def __add__(self, other):
        if not isinstance(other, Interval):
            return self + Interval(float(other), float(other))
        # pylint: disable=protected-access
        return Interval(
            down(lambda: self._lower + other._lower),
            up(lambda: self._upper + other._upper),
        )

    def __radd__(self, other):
        return self + other

    def __sub__(self, other):
        return self + (-other)

    def __rsub__(self, other):
        return (-self) + other

    def __mul__(self, other):
        if not isinstance(other, Interval):
            return self * Interval(float(other), float(other))
        if self.is_zero() or other.is_zero():
            return self.zero()
        if self is other:
            # equivalent to ** 2
            return self.__pow__(2)
        # pylint: disable=protected-access
        sl = self._lower
        su = self._upper
        ol = other._lower
        ou = other._upper
        return Interval(
            down(lambda: min_(sl*ol, sl*ou, su*ol, su*ou)),
            up(lambda: max_(sl*ol, sl*ou, su*ol, su*ou)),
        )

    def __rmul__(self, other):
        return self * other

    def __div__(self, other):
        if not isinstance(other, Interval):
            return self / Interval(float(other), float(other))
        if 0 in other:
            return Interval(None, None)
        return self * other.inverse()

    __truediv__ = __div__

    def __rdiv__(self, other):
        if not isinstance(other, Interval):
            return Interval(float(other), float(other)) / self
        if 0 in self:
            return Interval(None, None)
        return self.inverse() * other

    __rtruediv__ = __rdiv__

    def __abs__(self):
        if almostgte(self.lower_bound, 0):
            return self
        if almostlte(self.upper_bound, 0):
            return -self
        return Interval(0, up(lambda: max_(-self.lower_bound, self.upper_bound)))

    def __pow__(self, pow_):
        # Rules and tests from Boost::interval
        if isinstance(pow_, Interval):
            if not almosteq(pow_.lower_bound, pow_.upper_bound):
                return Interval(None, None)
            pow_ = pow_.lower_bound
            return self.__pow__(pow_)
        if not almosteq(pow_, int(pow_)):
            return Interval(None, None)
        pow_ = int(pow_)

        if almosteq(pow_, 0.0):
            return Interval(1, 1)

        if pow_ < 0:
            return 1.0 / self.__pow__(-pow_)

        is_odd = (abs(pow_) % 2) == 1
        if self.upper_bound < 0:
            # [-2, -1]
            yl = power(-self.upper_bound, pow_, RM.RD)
            yu = power(-self.lower_bound, pow_, RM.RU)
            if is_odd:
                return Interval(-yu, -yl)
            else:
                return Interval(yl, yu)
        elif self.lower_bound < 0:
            # [-1, 1]
            if is_odd:
                yl = power(-self.lower_bound, pow_, RM.RD)
                yu = power(self.upper_bound, pow_, RM.RU)
                return Interval(-yl, yu)
            else:
                return Interval(0, power(max(-self.lower_bound, self.upper_bound), pow_, RM.RU))
        else:
            # [1, 2]
            return Interval(power(self.lower_bound, pow_, RM.RD), power(self.upper_bound, pow_, RM.RU))

    def __eq__(self, other):
        if isinstance(other, (int, float)):
            return self.__eq__(Interval(other, other))

        if not isinstance(other, Interval):
            return False
        # pylint: disable=protected-access
        return (
            almosteq(self._lower, other._lower) and
            almosteq(self._upper, other._upper)
        )

    def __contains__(self, other):
        # pylint: disable=protected-access
        if isinstance(other, Interval):
            return (
                almostlte(self._lower, other._lower) and
                almostlte(other._upper, self._upper)
            )
        other = float(other)
        return almostlte(self._lower, other) and almostlte(other, self._upper)

    def __repr__(self):
        return '<{} at {}>'.format(str(self), hex(id(self)))

    def __str__(self):
        return 'Interval([{}, {}])'.format(self._lower, self._upper)

    def abs(self):
        return abs(self)

    @monotonic_increasing
    def _sqrt(self):
        return sqrt

    def _sqr(self):
        new_lower = down(lambda: self._lower * self._lower)
        new_upper = up(lambda: self._upper * self._upper)
        return Interval(new_lower, new_upper)

    @monotonic_increasing
    def _exp(self):
        return exp

    @monotonic_increasing
    def _log(self):
        return log

    def _sin(self):
        if almostgte(self.size(), 2*pi):
            return Interval(None, None)
        lower = self._lower % (2*pi)
        upper = lower + (self._upper - self._lower)
        new_lower = min_(sin(lower, RM.RD), sin(upper, RM.RD))
        new_upper = max_(sin(lower, RM.RU), sin(upper, RM.RU))
        new = Interval(new_lower, new_upper)
        translated_interval = Interval(lower, upper)
        if 0.5 * pi in translated_interval:
            new_upper = 1.0
        if 1.5 * pi in translated_interval:
            new_lower = -1.0
        return Interval(new_lower, new_upper)

    def _cos(self):
        if almostgte(self.size(), 2*pi):
            return Interval(None, None)
        # translate by pi/2 and class sin
        pi_2 = pi/2.0
        return (self + pi_2).sin()

    def _tan(self):
        if almostgte(self.size(), pi):
            return Interval(None, None)
        lower = self._lower % pi
        upper = lower  + (self._upper - self._lower)
        new_lower = min_(tan(lower, RM.RD), tan(upper, RM.RD))
        new_upper = max_(tan(lower, RM.RU), tan(upper, RM.RU))

        if almosteq(lower, 0.5 * pi):
            new_lower = None

        if almosteq(upper, 0.5 * pi):
            new_upper = None

        if new_lower is not None and new_upper is not None:
            new = Interval(new_lower, new_upper)
            if 0.5*pi in new or pi in new:
                return Interval(None, None)
        return Interval(new_lower, new_upper)

    @monotonic_increasing
    def _asin(self):
        return asin

    @monotonic_decreasing
    def _acos(self):
        return acos

    @monotonic_increasing
    def _atan(self):
        return atan


class EmptyInterval(Interval):
    """An empty interval."""
    def __init__(self):
        super().__init__(0, 0)

    def __str__(self):
        return 'EmptyInterval([{}, {}])'.format(self._lower, self._upper)



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Convexity detection rules for product expressions."""
from suspect.pyomo.expressions import MonomialTermExpression, nonpyomo_leaf_types
from suspect.convexity.convexity import Convexity
from suspect.convexity.rules.rule import ConvexityRule
from suspect.expression import ExpressionType


class ProductRule(ConvexityRule):
    """Return convexity of product."""
    def apply(self, expr, convexity, monotonicity, bounds):
        if isinstance(expr, MonomialTermExpression):
            _, var = expr.args
            return convexity[var]

        f, g = expr.args
        if f is g:
            return _product_as_square_convexity(f, convexity, bounds)

        # -c*x*x is encoded as Linear([-c], [x])*Variable(x)
        g_is_var = type(g) not in nonpyomo_leaf_types and g.is_variable_type()
        f_is_var = type(f) not in nonpyomo_leaf_types and f.is_variable_type()

        if isinstance(f, MonomialTermExpression) and g_is_var:
            cvx = _product_monomial_term_by_variable_convexity(f, g)
            if cvx is not None:
                return cvx

        if f_is_var and isinstance(g, MonomialTermExpression):
            cvx = _product_monomial_term_by_variable_convexity(g, f)
            if cvx is not None:
                return cvx

        mono_g = monotonicity[g]
        mono_f = monotonicity[f]
        if mono_g.is_constant():
            return _product_convexity(f, g, convexity, bounds)
        elif mono_f.is_constant():
            return _product_convexity(g, f, convexity, bounds)
        return Convexity.Unknown


def _product_as_square_convexity(f, convexity, bounds):
    # same as f**2
    cvx_f = convexity[f]
    bounds_f = bounds[f]
    if cvx_f.is_linear():
        return Convexity.Convex
    elif cvx_f.is_convex() and bounds_f.is_nonnegative():
        return Convexity.Convex
    elif cvx_f.is_concave() and bounds_f.is_nonpositive():
        return Convexity.Convex
    return Convexity.Unknown


def _product_monomial_term_by_variable_convexity(f, g):
    coef, f_v = f.args
    if f_v is g:
        if coef > 0:
            return Convexity.Convex
        return Convexity.Concave
    return None


def _product_convexity(f, g, convexity, bounds):
    cvx_f = convexity[f]
    bounds_g = bounds[g]

    if cvx_f.is_linear():
        return cvx_f
    elif cvx_f.is_convex() and bounds_g.is_nonnegative():
        return Convexity.Convex
    elif cvx_f.is_concave() and bounds_g.is_nonpositive():
        return Convexity.Convex
    elif cvx_f.is_concave() and bounds_g.is_nonnegative():
        return Convexity.Concave
    elif cvx_f.is_convex() and bounds_g.is_nonpositive():
        return Convexity.Concave
    return Convexity.Unknown



# pylint: skip-file
import pytest
from suspect.math import RoundMode as RM
from suspect.math.arbitrary_precision import *


@pytest.mark.parametrize('func', [sqrt, log, exp, sin])
def test_func(func):
    assert func(make_number(123.4), RM.RN) == func(make_number(123.4), RM.RD)
    assert func(make_number(123.4), RM.RU) == func(make_number(123.4), RM.RD)
    assert func(make_number(123.4), RM.RZ) == func(make_number(123.4), RM.RD)


class TestAlmostEq(object):
    def test_inf(self):
        assert almosteq(inf, inf)
        assert not almosteq(inf, -inf)
        assert almosteq(-inf, -inf)

    def test_finite(self):
        assert almosteq(pi, pi)
        assert not almosteq(2e20, inf)


class TestAlomstGte(object):
    def test_inf(self):
        assert almostgte(inf, -inf)
        assert almostgte(inf, inf)
        assert almostgte(-inf, -inf)

    def test_finite(self):
        assert almostgte(pi, pi)


class TestAlomstLte(object):
    def test_inf(self):
        assert almostlte(-inf, inf)
        assert almostlte(inf, inf)
        assert almostlte(-inf, -inf)

    def test_finite(self):
        assert almostlte(pi, pi)



# Copyright 2019 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name
"""Floating point arithmetic."""
import numpy as np

make_number = np.float
inf = np.inf
zero = float(0.0)
pi = np.pi
isnan = lambda n: np.isnan(n)
isinf = lambda n: np.isinf(n)


def _declare_unary_function(name, fun):
    # skip rounding mode since we don't round
    globals()[name] = lambda n, _: float(fun(n))


def _declare_binary_function(name, fun):
    # skip rounding mode since we don't round
    globals()[name] = lambda x, y, _: float(fun(x, y))


_UNARY_FUNCTIONS = [
    ('sqrt', 'sqrt'), ('log', 'log'), ('exp', 'exp'), ('sin', 'sin'),
    ('asin', 'arcsin'), ('cos', 'cos'), ('acos', 'arccos'),
    ('tan', 'tan'), ('atan', 'arctan')
]
for fun_name, fun in _UNARY_FUNCTIONS:
    _declare_unary_function(fun_name, getattr(np, fun))


_BINARY_FUNCTIONS = ['power']
for fun in _BINARY_FUNCTIONS:
    _declare_binary_function(fun, getattr(np, fun))


def down(f):
    """Perform computation rounding down."""
    return f()


def up(f):
    """Perform computation rounding down."""
    return f()


def min_(*args):
    """Returns the minimum."""
    return min(a for a in args if not isnan(a))


def max_(*args):
    """Return the maximum."""
    return max(a for a in args if not isnan(a))


def almosteq(a, b, rel_eps=None, abs_eps=None):
    """Floating point equality check between `a` and `b`."""
    # inf != inf, but we want inf == inf
    if rel_eps is None:
        rel_eps = 1.0e-5
    if abs_eps is None:
        abs_eps = 1.0e-8
    return np.all(np.isclose(a, b, rtol=rel_eps, atol=abs_eps))


def almostgte(a, b):
    """Return True if a >= b."""
    return np.all(np.logical_or(a > b, almosteq(a, b)))


def almostlte(a, b):
    """Return True if a <= b."""
    return np.all(np.logical_or(a < b, almosteq(a, b)))
